from sklearn.svm import LinearSVC as SklearnSVC

from immuneML.ml_methods.classifiers.SklearnMethod import SklearnMethod
from scripts.specification_util import update_docs_per_mapping


class SVC(SklearnMethod):
    """
    This is a wrapper of scikit-learn’s LinearSVC class. Please see the
    `LinearSVC scikit-learn documentation <https://scikit-learn.org/stable/modules/generated/sklearn.svm.LinearSVC.html>`_
    of SVC for the parameters.

    Note: if you are interested in plotting the coefficients of the SVC model,
    consider running the :ref:`Coefficients` report.

    For usage instructions, check :py:obj:`~immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod`.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_svc: # user-defined method name
                    SVC: # name of the ML method
                        # sklearn parameters (same names as in original sklearn class)
                        C: [0.01, 0.1, 1, 10, 100] # find the optimal value for C
                        # Additional parameter that determines whether to print convergence warnings
                        show_warnings: True
                    # if any of the parameters under SVC is a list and model_selection_cv is True,
                    # a grid search will be done over the given parameters, using the number of folds specified in model_selection_n_folds,
                    # and the optimal model will be selected
                    model_selection_cv: True
                    model_selection_n_folds: 5
                # alternative way to define ML method with default values:
                my_default_svc: SVC

    """

    def __init__(self, parameter_grid: dict = None, parameters: dict = None):
        _parameter_grid = parameter_grid if parameter_grid is not None else {}
        _parameters = parameters if parameters is not None else {}

        super(SVC, self).__init__(parameter_grid=_parameter_grid, parameters=_parameters)

    def _get_ml_model(self, cores_for_training: int = 2, X=None):
        return SklearnSVC(**self._parameters)

    def can_predict_proba(self) -> bool:
        return False

    def can_fit_with_example_weights(self) -> bool:
        return True

    def get_params(self, for_refitting=False):
        params = self.model.get_params()
        if not for_refitting:
            params["coefficients"] = self.model.coef_[0].tolist()
            params["intercept"] = self.model.intercept_.tolist()
        return params

    @staticmethod
    def get_documentation():
        doc = str(SVC.__doc__)

        mapping = {
            "For usage instructions, check :py:obj:`~immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod`.": SklearnMethod.get_usage_documentation("SVC"),
        }

        doc = update_docs_per_mapping(doc, mapping)
        return doc



import hashlib
import logging
from pathlib import Path

import numpy as np
import torch
import yaml
from sklearn.exceptions import NotFittedError
from sklearn.model_selection import train_test_split

from immuneML.data_model.EncodedData import EncodedData
from immuneML.encodings.deeprc.DeepRCEncoder import DeepRCEncoder
from immuneML.environment.Label import Label
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_methods.util.Util import Util
from immuneML.util.FilenameHandler import FilenameHandler
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.ReflectionHandler import ReflectionHandler


class DeepRC(MLMethod):
    """
    This classifier uses the DeepRC method for repertoire classification. The DeepRC ML method should be used in combination
    with the DeepRC encoder. Also consider using the :ref:`DeepRCMotifDiscovery` report for interpretability.

    Notes:

    - DeepRC uses PyTorch functionalities that depend on GPU. Therefore, DeepRC does not work on a CPU.

    - This wrapper around DeepRC currently only supports binary classification.

    Reference:
    Michael Widrich, Bernhard Schäfl, Milena Pavlović, Geir Kjetil Sandve, Sepp Hochreiter, Victor Greiff, Günter Klambauer
    ‘DeepRC: Immune repertoire classification with attention-based deep massive multiple instance learning’.
    bioRxiv preprint doi: `https://doi.org/10.1101/2020.04.12.038158 <https://doi.org/10.1101/2020.04.12.038158>`_


    **Specification arguments:**

    - validation_part (float):  the part of the data that will be used for validation, the rest will be used for training.

    - add_positional_information (bool): whether positional information should be included in the input features.

    - kernel_size (int): the size of the 1D-CNN kernels.

    - n_kernels (int): the number of 1D-CNN kernels in each layer.

    - n_additional_convs (int): Number of additional 1D-CNN layers after first layer

    - n_attention_network_layers (int): Number of attention layers to compute keys

    - n_attention_network_units (int): Number of units in each attention layer

    - n_output_network_units (int): Number of units in the output layer

    - consider_seq_counts (bool): whether the input data should be scaled by the receptor sequence counts.

    - sequence_reduction_fraction (float): Fraction of number of sequences to which to reduce the number of sequences per bag based on attention weights. Has to be in range [0,1].

    - reduction_mb_size (int): Reduction of sequences per bag is performed using minibatches of reduction_mb_size` sequences to compute the attention weights.

    - n_updates (int): Number of updates to train for

    - n_torch_threads (int):  Number of parallel threads to allow PyTorch

    - learning_rate (float): Learning rate for adam optimizer

    - l1_weight_decay (float): l1 weight decay factor. l1 weight penalty will be added to loss, scaled by `l1_weight_decay`

    - l2_weight_decay (float): l2 weight decay factor. l2 weight penalty will be added to loss, scaled by `l2_weight_decay`

    - sequence_counts_scaling_fn: it can either be `log` (logarithmic scaling of sequence counts) or None

    - sequence_counts_scaling_fn: it can either be `log` (logarithmic scaling of sequence counts) or None

    - evaluate_at (int): Evaluate model on training and validation set every `evaluate_at` updates. This will also check for a new best model for early stopping.

    - sample_n_sequences (int): Optional random sub-sampling of `sample_n_sequences` sequences per repertoire. Number of sequences per repertoire might be smaller than `sample_n_sequences` if repertoire is smaller or random indices have been drawn multiple times. If None, all sequences will be loaded for each repertoire.

    - training_batch_size (int): Number of repertoires per minibatch during training.

    - n_workers (int): Number of background processes to use for converting dataset to hdf5 container and training set data loader.

    - pytorch_device_name (str): The name of the pytorch device to use. This name will be passed to  torch.device(self.pytorch_device_name). The default value is cuda:0


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_deeprc_method:
                    DeepRC:
                        validation_part: 0.2
                        add_positional_information: True
                        kernel_size: 9

    """

    def __init__(self, validation_part, add_positional_information, kernel_size, n_kernels,
                 n_additional_convs, n_attention_network_layers, n_attention_network_units, n_output_network_units,
                 consider_seq_counts, sequence_reduction_fraction, reduction_mb_size, n_updates, n_torch_threads,
                 learning_rate, l1_weight_decay, l2_weight_decay, evaluate_at, sample_n_sequences, training_batch_size,
                 n_workers, sequence_counts_scaling_fn, keep_dataset_in_ram, pytorch_device_name):
        super(DeepRC, self).__init__()

        if not ReflectionHandler.is_installed("deeprc"):
            raise RuntimeError(f"{DeepRC.__name__}: deeprc module is not installed. Please check the documentation at "
                               f"https://docs.immuneml.uio.no/installation/install_with_package_manager.html for "
                               f"instructions how to install it.")

        from deeprc.training import train
        self.training_function = train

        self.model = None
        self.max_seq_len = None

        self.label = None
        self.feature_names = None

        self.keep_dataset_in_ram = keep_dataset_in_ram
        self.pytorch_device_name = pytorch_device_name
        self.pytorch_device = torch.device(self.pytorch_device_name)

        # ML model setting (not inherited from DeepRC code)
        self.validation_part = validation_part

        # DeepRC class settings:
        self.add_positional_information = add_positional_information
        self.n_input_features = 20
        self.kernel_size = kernel_size
        self.n_kernels = n_kernels
        self.n_additional_convs = n_additional_convs
        self.n_attention_network_layers = n_attention_network_layers
        self.n_attention_network_units = n_attention_network_units
        self.n_output_network_units = n_output_network_units
        self.consider_seq_counts = consider_seq_counts
        self.sequence_reduction_fraction = sequence_reduction_fraction
        self.reduction_mb_size = reduction_mb_size

        from deeprc.dataset_readers import log_sequence_count_scaling, no_sequence_count_scaling
        self.sequence_counts_scaling_fn = log_sequence_count_scaling if sequence_counts_scaling_fn == "log" \
            else no_sequence_count_scaling

        # train function settings:
        self.evaluate_at = evaluate_at
        self.n_updates = n_updates
        self.n_torch_threads = n_torch_threads
        self.learning_rate = learning_rate
        self.l1_weight_decay = l1_weight_decay
        self.l2_weight_decay = l2_weight_decay

        # Dataloader related settings:
        self.sample_n_sequences = sample_n_sequences
        self.training_batch_size = training_batch_size
        self.n_workers = n_workers

    def _metadata_to_hdf5(self, metadata_filepath: Path, label_name: str):
        from deeprc.dataset_converters import DatasetToHDF5

        hdf5_filepath = metadata_filepath.parent / f"{metadata_filepath.stem}.hdf5"
        converter = DatasetToHDF5(repertoiresdata_directory=str(metadata_filepath.parent),
                                  sequence_column=DeepRCEncoder.SEQUENCE_COLUMN,
                                  sequence_counts_column=DeepRCEncoder.COUNTS_COLUMN,
                                  column_sep=DeepRCEncoder.SEP, filename_extension=f".{DeepRCEncoder.EXTENSION}",
                                  verbose=False)
        converter.save_data_to_file(output_file=str(hdf5_filepath), n_workers=self.n_workers)

        return hdf5_filepath

    def _load_dataset_in_ram(self, hdf5_filepath: Path):
        import h5py

        with h5py.File(str(hdf5_filepath), 'r') as hf:
            pre_loaded_hdf5_file = dict()
            pre_loaded_hdf5_file['seq_lens'] = hf['sampledata']['seq_lens'][:]
            pre_loaded_hdf5_file['counts_per_sequence'] = hf['sampledata']['counts_per_sequence'][:]
            pre_loaded_hdf5_file['amino_acid_sequences'] = hf['sampledata']['amino_acid_sequences'][:]

        return pre_loaded_hdf5_file

    def _get_train_val_indices(self, n_examples, classes):
        """splits the data to training and validation and attempts to preserve the class distribution if possible"""

        indices = np.arange(0, n_examples)
        train_indices, val_indices = train_test_split(indices, test_size=self.validation_part, shuffle=True,
                                                      stratify=classes)

        return train_indices, val_indices

    def make_data_loader(self, full_dataset, indices, label_name, eval_only: bool, is_train: bool, n_workers=1):
        """
        Creates a pytorch dataloader using DeepRC's RepertoireDataReaderBinary

        :param hdf5_filepath: the path to the HDF5 file
        :param pre_loaded_hdf5_file: Optional: It is faster to load the hdf5 file into the RAM as dictionary instead
            of keeping it on the disk. `pre_loaded_hdf5_file` is the loaded hdf5 file as dictionary.
            If None, the hdf5 file will be read from the disk and consume less RAM.
        :param indices: indices of the subset of repertoires in the data that will be used for this dataset.
                If 'None', all repertoires will be used.
        :param label_name: the name of the label to be predicted
        :param eval_only: whether the dataloader will only be used for evaluation (no training).
                if false, sample_n_sequences can be set
        :param is_train: whether this is a dataloader for training data. If true, self.training_batch_size is used.
        :param n_workers: the number of workers used in torch.utils.data.DataLoader
        :return: a Pytorch dataloader
        """
        from deeprc.dataset_readers import no_stack_collate_fn
        from deeprc.dataset_readers import RepertoireDatasetSubset as DeepRCRepDatasetSubset
        from torch.utils.data import DataLoader

        sample_n_sequences = None if eval_only else self.sample_n_sequences
        batch_size = self.training_batch_size if is_train else 1

        dataset = DeepRCRepDatasetSubset(dataset=full_dataset, indices=indices,
                                         sample_n_sequences=sample_n_sequences)

        data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=True, num_workers=n_workers,
                                 collate_fn=no_stack_collate_fn, multiprocessing_context='spawn')
        return data_loader

    def _prepare_caching_params(self, encoded_data: EncodedData, type: str, label_name: str):
        return (("metadata_filepath", str(encoded_data.info["metadata_filepath"])),
                ("y", hashlib.sha256(str(encoded_data.labels[label_name]).encode("utf-8")).hexdigest()),
                ("label_name", label_name),
                ("type", type),
                ("validation_part", self.validation_part),
                ("add_positional_information", self.add_positional_information),
                ("n_input_features", self.n_input_features),
                ("kernel_size", self.kernel_size),
                ("n_kernels", self.n_kernels),
                ("n_additional_convs", self.n_additional_convs),
                ("n_attention_network_layers", self.n_attention_network_layers),
                ("n_attention_network_units", self.n_attention_network_units),
                ("n_output_network_units", self.n_output_network_units),
                ("consider_seq_counts", self.consider_seq_counts),
                ("sequence_reduction_fraction", self.sequence_reduction_fraction),
                ("reduction_mb_size", self.reduction_mb_size),
                ("n_updates", self.n_updates),
                ("n_torch_threads", self.n_torch_threads),
                ("learning_rate", self.learning_rate),
                ("l1_weight_decay", self.l1_weight_decay),
                ("l2_weight_decay", self.l2_weight_decay),
                ("sample_n_sequences", self.sample_n_sequences),
                ("training_batch_size", self.training_batch_size),
                ("n_workers", self.n_workers),
                ("evaluate_at", self.evaluate_at),
                ("pytorch_device_name", self.pytorch_device_name))

    def _fit(self, encoded_data: EncodedData, cores_for_training: int = 2):

        hdf5_filepath, pre_loaded_hdf5_file = self._convert_dataset_to_hdf5(encoded_data, self.label)

        train_indices, val_indices = self._get_train_val_indices(len(encoded_data.example_ids),
                                                                 encoded_data.labels[self.label.name])
        self.max_seq_len = encoded_data.info["max_sequence_length"]

        self._fit_for_label(encoded_data.info["metadata_filepath"], hdf5_filepath, train_indices, val_indices,
                            self.label,
                            cores_for_training)

        return self.model

    def _convert_dataset_to_hdf5(self, encoded_data, label):

        hdf5_filepath = self._metadata_to_hdf5(encoded_data.info["metadata_filepath"], label.name)
        pre_loaded_hdf5_file = None  # self._load_dataset_in_ram(hdf5_filepath) if self.keep_dataset_in_ram else None

        return hdf5_filepath, pre_loaded_hdf5_file

    def _make_model(self, task_definition):
        from deeprc.architectures import DeepRC as DeepRCInternal

        # Create sequence embedding network (for CNN, kernel_size and n_kernels are important hyperparameters)
        from deeprc.architectures import SequenceEmbeddingCNN
        sequence_embedding_network = SequenceEmbeddingCNN(
            n_input_features=self.n_input_features + 3 * self.add_positional_information,
            kernel_size=self.kernel_size,
            n_kernels=self.n_kernels, n_layers=1)

        # Create attention network
        from deeprc.architectures import AttentionNetwork
        attention_network = AttentionNetwork(n_input_features=self.n_kernels, n_layers=self.n_attention_network_layers,
                                             n_units=self.n_attention_network_units)

        # Create output network
        from deeprc.architectures import OutputNetwork
        output_network = OutputNetwork(n_input_features=self.n_kernels,
                                       n_output_features=task_definition.get_n_output_features(), n_layers=1,
                                       n_units=self.n_output_network_units)
        # Combine networks to DeepRC network
        return DeepRCInternal(max_seq_len=self.max_seq_len, sequence_embedding_network=sequence_embedding_network,
                              attention_network=attention_network, output_network=output_network,
                              consider_seq_counts=False, n_input_features=self.n_input_features,
                              add_positional_information=self.add_positional_information,
                              sequence_reduction_fraction=self.sequence_reduction_fraction,
                              reduction_mb_size=self.reduction_mb_size,
                              device=self.pytorch_device).to(device=self.pytorch_device)

    def _make_task_definition(self, label):
        from deeprc.task_definitions import TaskDefinition, BinaryTarget, MulticlassTarget

        if len(label.values) == 2:
            target = BinaryTarget(column_name=label.name, true_class_value=label.positive_class)
        else:
            target = MulticlassTarget(column_name=label.name, possible_target_values=label.values)

        return TaskDefinition(targets=[target])

    def _fit_for_label(self, metadata_file, hdf5_filepath: Path, train_indices, val_indices,
                       label: Label, cores_for_training: int):

        from deeprc.dataset_readers import RepertoireDataset as DeepRCRepDataset

        task_definition = self._make_task_definition(label)

        full_dataset = DeepRCRepDataset(metadata_filepath=metadata_file, hdf5_filepath=str(hdf5_filepath),
                                        sample_id_column=DeepRCEncoder.ID_COLUMN,
                                        metadata_file_column_sep=",",
                                        task_definition=task_definition, keep_in_ram=self.keep_dataset_in_ram,
                                        inputformat='NCL',
                                        sequence_counts_scaling_fn=self.sequence_counts_scaling_fn)

        train_dataloader = self.make_data_loader(full_dataset, train_indices, label.name, eval_only=False,
                                                 is_train=True, n_workers=self.n_workers)

        train_eval_dataloader = self.make_data_loader(full_dataset, train_indices, eval_only=True, is_train=True,
                                                      n_workers=1, label_name=label.name)

        val_dataloader = self.make_data_loader(full_dataset, val_indices, eval_only=True, is_train=False,
                                               label_name=label.name, n_workers=1)

        self.model = self._make_model(task_definition)

        self.training_function(self.model, trainingset_dataloader=train_dataloader,
                               trainingset_eval_dataloader=train_eval_dataloader,
                               validationset_eval_dataloader=val_dataloader,
                               results_directory=self.result_path / "deep_rc_log",
                               n_updates=self.n_updates, num_torch_threads=self.n_torch_threads,
                               learning_rate=self.learning_rate,
                               l1_weight_decay=self.l1_weight_decay, l2_weight_decay=self.l2_weight_decay,
                               show_progress=False, device=self.pytorch_device, evaluate_at=self.evaluate_at,
                               task_definition=task_definition, early_stopping_target_id=label.name)

    def fit_by_cross_validation(self, encoded_data: EncodedData, number_of_splits: int = 5, label: Label = None,
                                cores_for_training: int = -1,
                                optimization_metric=None):
        logging.warning("DeepRC: cross-validation on this classifier is not defined: fitting one model instead...")
        self.fit(encoded_data, label)

    def get_params(self, for_refitting=False):
        return {name: param.data.tolist() for name, param in self.model.named_parameters()}

    def check_is_fitted(self, label_name: str):
        if label_name != self.label.name:
            raise NotFittedError("This DeepRCs instance is not fitted yet. "
                                 "Call 'fit' with appropriate arguments before using this method.")

    def _predict(self, encoded_data: EncodedData):
        probabilities = self._predict_proba(encoded_data)

        pos_class_probs = probabilities[self.label.name][self.label.positive_class]
        negative_class = self.label.get_binary_negative_class()

        # TODO: check what is returned here and how it works with multiclass
        return {self.label.name: [self.label.positive_class if probability > 0.5 else negative_class for probability in
                                  pos_class_probs]}

    def _predict_proba(self, encoded_data: EncodedData):
        from deeprc.dataset_readers import RepertoireDataset as DeepRCRepDataset
        self.check_is_fitted(self.label.name)

        hdf5_filepath, _ = self._convert_dataset_to_hdf5(encoded_data, self.label)
        task_definition = self._make_task_definition(self.label)

        test_dataset = DeepRCRepDataset(metadata_filepath=encoded_data.info['metadata_filepath'],
                                        hdf5_filepath=str(hdf5_filepath),
                                        sample_id_column=DeepRCEncoder.ID_COLUMN,
                                        metadata_file_column_sep=DeepRCEncoder.SEP,
                                        task_definition=task_definition, keep_in_ram=self.keep_dataset_in_ram,
                                        inputformat='NCL',
                                        sequence_counts_scaling_fn=self.sequence_counts_scaling_fn)

        test_dataloader = self.make_data_loader(test_dataset, indices=None, label_name=self.label.name, eval_only=True,
                                                is_train=False)

        probs_pos_class = self._model_predict(self.model, test_dataloader)

        # TODO: update for multiclass
        return {self.label.name: {self.label.positive_class: probs_pos_class,
                                  self.label.get_binary_negative_class(): 1 - probs_pos_class}}

    def _model_predict(self, model, dataloader):
        """Based on the DeepRC function evaluate (deeprc.training.evaluate)"""
        from tqdm import tqdm

        with torch.no_grad():
            model.to(device=self.pytorch_device)
            scoring_predictions = []
            for scoring_data in tqdm(dataloader, total=len(dataloader), desc="Evaluating model",
                                     disable=True, position=1):
                # Get samples as lists
                labels, inputs, sequence_lengths, counts_per_sequence, sample_ids = scoring_data

                # Apply attention-based sequence reduction and create minibatch
                labels, inputs, sequence_lengths, n_sequences = model.reduce_and_stack_minibatch(
                    labels, inputs, sequence_lengths, counts_per_sequence)

                # Compute predictions from reduced sequences
                logit_outputs = model(inputs, n_sequences)
                prediction = torch.sigmoid(logit_outputs)
                scoring_predictions.append(prediction)

            # predictions
            scoring_predictions = torch.cat(scoring_predictions, dim=0).float().cpu().numpy()

        return scoring_predictions

    def load(self, path: Path):
        name = FilenameHandler.get_filename(self.__class__.__name__, "pt")
        file_path = path / name
        if file_path.is_file():
            self.model = torch.load(str(file_path))
            self.model.eval()
        else:
            raise FileNotFoundError(f"{self.__class__.__name__} model could not be loaded from {file_path}. "
                                    f"Check if the path to the {name} file is properly set.")

        params_path = path / FilenameHandler.get_filename(self.__class__.__name__, "yaml")

        if params_path.is_file():
            with params_path.open("r") as file:
                desc = yaml.safe_load(file)
                if "label" in desc:
                    setattr(self, "label", Label(**desc["label"]))
                for param in ["feature_names", "classes"]:
                    if param in desc:
                        setattr(self, param, desc[param])

    def store(self, path):
        PathBuilder.build(path)
        name = FilenameHandler.get_filename(self.__class__.__name__, "pt")
        torch.save(self.model, str(path / name))

        params_path = path / FilenameHandler.get_filename(self.__class__.__name__, "yaml")

        with params_path.open("w") as file:
            desc = {
                **(self.get_params()),
                "feature_names": self.get_feature_names(),
                "classes": self.get_classes()
            }
            if self.label is not None:
                desc["label"] = self.label.get_desc_for_storage()

            yaml.dump(desc, file)

    def get_package_info(self) -> str:
        return Util.get_immuneML_version()

    def can_predict_proba(self) -> bool:
        return True

    def can_fit_with_example_weights(self) -> bool:
        return False

    def get_compatible_encoders(self):
        return [DeepRCEncoder]



from pathlib import Path

import math
import pandas as pd
from plotly.subplots import make_subplots
import plotly.graph_objects as go
import plotly.express as px

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class LabelDist(DataReport):
    """
    LabelDist report plots the distribution of label values for all labels provided as
    input to the report.

    Specification arguments:

    - labels (list): list of label names as they appear in the metadata file (RepertoireDataset)
      or in data files (Receptor/SequenceDataset).

    YAML specification:

    .. code-block: yaml

        reports:
            label_count_report:
                LabelCount:
                    labels: ['diagnosis', 'age_group', 'batch']
    """

    def __init__(self, dataset: Dataset = None, result_path: Path = None, name: str = None,
                 number_of_processes: int = 1, labels: list = None):
        super().__init__(dataset=dataset, result_path=result_path, name=name, number_of_processes=number_of_processes)
        self.labels = labels

    @classmethod
    def build_object(cls, **kwargs):
        ParameterValidator.assert_keys(list(kwargs.keys()), ["labels", 'name'], "LabelDist report", "LabelCount")
        ParameterValidator.assert_all_type_and_value(kwargs["labels"], str, "LabelDist", "labels")
        return cls(name=kwargs["name"], labels=kwargs["labels"])

    def _generate(self) -> ReportResult:

        df = self.dataset.get_metadata(self.labels, return_df=True)

        n_cols = 2
        n_rows = math.ceil(len(df.columns) / n_cols)

        fig = make_subplots(rows=n_rows, cols=n_cols, subplot_titles=df.columns)

        colors = px.colors.qualitative.Vivid * math.ceil(len(df.columns) / len(px.colors.qualitative.Vivid))

        for i, col in enumerate(df.columns):
            row = i // n_cols + 1
            col_pos = i % n_cols + 1
            color = colors[i]

            if pd.api.types.is_numeric_dtype(df[col]):
                trace = go.Histogram(x=df[col], name=col, marker_color=color)
            else:
                counts = df[col].value_counts()
                trace = go.Bar(x=counts.index.astype(str), y=counts.values, name=col, marker_color=color)

            fig.add_trace(trace, row=row, col=col_pos)

        fig.update_layout(template='plotly_white', height=300 * n_rows, showlegend=False,
                          title_text="Label distributions")

        path = PathBuilder.build(self.result_path) / f"{self.name}_label_distributions.html"
        PlotlyUtil.write_image_to_file(fig, path, df.shape[0])

        return ReportResult(name=self.name, info='Label distributions',
                            output_figures=[ReportOutput(name='Label distributions', path=path)])




from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.simulation.dataset_generation.RandomDatasetGenerator import RandomDatasetGenerator
from immuneML.util.ParameterValidator import ParameterValidator


class RandomSequenceDatasetImport(DataImport):
    """
    Returns a SequenceDataset consisting of randomly generated sequences, which can be used for benchmarking purposes.
    The sequences consist of uniformly chosen amino acids or nucleotides.


    **Specification arguments:**

    - sequence_count (int): The number of sequences the SequenceDataset should contain.

    - length_probabilities (dict): A mapping where the keys correspond to different sequence lengths and the values
      are the probabilities for choosing each sequence length. For example, to create a random SequenceDataset where
      40% of the sequences would be of length 10, and 60% of the sequences would have length 12, this mapping would
      need to be specified:

    .. indent with spaces
    .. code-block:: yaml

            10: 0.4
            12: 0.6

    - labels (dict): A mapping that specifies randomly chosen labels to be assigned to the sequences. One or multiple
      labels can be specified here. The keys of this mapping are the labels, and the values consist of another mapping
      between label classes and their probabilities. For example, to create a random SequenceDataset with the label
      cmv_epitope where 70% of the sequences has class binding and the remaining 30% has class not_binding, the
      following mapping should be specified:

    .. indent with spaces
    .. code-block:: yaml

            cmv_epitope:
                binding: 0.7
                not_binding: 0.3

    - region_type (str): which region_type to assign to all randomly generated sequences


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            datasets:
                my_random_dataset:
                    format: RandomSequenceDataset
                    params:
                        sequence_count: 100 # number of random sequences to generate
                        length_probabilities:
                            14: 0.8 # 80% of all generated sequences for all sequences will have length 14
                            15: 0.2 # 20% of all generated sequences across all sequences will have length 15
                        labels:
                            epitope1: # label name
                                True: 0.5 # 50% of the sequences will have class True
                                False: 0.5 # 50% of the sequences will have class False
                            epitope2: # next label with classes that will be assigned to sequences independently of the previous label or other parameters
                                1: 0.3 # 30% of the generated sequences will have class 1
                                0: 0.7 # 70% of the generated sequences will have class 0

    """

    def __init__(self, params: dict, dataset_name: str):
        super().__init__({}, dataset_name)
        self.params = params
        self.dataset_name = dataset_name

    def import_dataset(self) -> SequenceDataset:
        """
        Returns randomly generated receptor dataset according to the parameters;

        **YAML specification:**

            result_path: path/where/to/store/results/
            sequence_count: 100 # number of random sequences to generate
            chain_1_length_probabilities:
                14: 0.8 # 80% of all generated sequences for all sequences will have length 14
                15: 0.2 # 20% of all generated sequences across all sequences will have length 15
            labels:
                epitope1: # label name
                    True: 0.5 # 50% of the sequences will have class True
                    False: 0.5 # 50% of the sequences will have class False
                epitope2: # next label with classes that will be assigned to sequences independently of the previous label or other parameters
                    1: 0.3 # 30% of the generated sequences will have class 1
                    0: 0.7 # 70% of the generated sequences will have class 0

        """
        valid_keys = ["sequence_count", "length_probabilities", "labels", "result_path", 'region_type']
        ParameterValidator.assert_all_in_valid_list(list(self.params.keys()), valid_keys, "RandomSequenceDatasetImport", "params")

        return RandomDatasetGenerator.generate_sequence_dataset(sequence_count=self.params["sequence_count"],
                                                                length_probabilities=self.params["length_probabilities"],
                                                                labels=self.params["labels"],
                                                                path=self.params["result_path"],
                                                                region_type=self.params['region_type'],
                                                                name=self.dataset_name)



import logging
from functools import partial
from multiprocessing.pool import Pool
from pathlib import Path

import pandas as pd
import numpy as np
from sklearn.metrics import precision_score, recall_score, confusion_matrix

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.data_model.EncodedData import EncodedData
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.motif_encoding.PositionalMotifParams import PositionalMotifParams
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.util.EncoderHelper import EncoderHelper
from immuneML.util.ParameterValidator import ParameterValidator

from immuneML.encodings.motif_encoding.PositionalMotifHelper import PositionalMotifHelper
from immuneML.util.PathBuilder import PathBuilder


class MotifEncoder(DatasetEncoder):
    """
    This encoder enumerates every possible positional motif in a sequence dataset, and keeps only the motifs associated with the positive class.
    A 'motif' is defined as a combination of position-specific amino acids. These motifs may contain one or multiple gaps.
    Motifs are filtered out based on a minimal precision and recall threshold for predicting the positive class.

    Note: the MotifEncoder can only be used for sequences of the same length.

    The ideal recall threshold(s) given a user-defined precision threshold can be calibrated using the
    :py:obj:`~immuneML.reports.data_reports.MotifGeneralizationAnalysis` report. It is recommended to first run this report
    in :py:obj:`~immuneML.workflows.instructions.exploratory_analysis.ExploratoryAnalysisInstruction` before using this encoder for ML.

    This encoder can be used in combination with the :py:obj:`~immuneML.ml_methods.BinaryFeatureClassifier` in order to
    learn a minimal set of compatible motifs for predicting the positive class.
    Alternatively, it may be combined with scikit-learn methods, such as for example :py:obj:`~immuneML.ml_methods.LogisticRegression`,
    to learn a weight per motif.

    **Dataset type:**

    - SequenceDatasets


    **Specification arguments:**

    - max_positions (int): The maximum motif size. This is number of positional amino acids the motif consists of (excluding gaps). The default value for max_positions is 4.

    - min_positions (int): The minimum motif size (see also: max_positions). The default value for max_positions is 1.

    - no_gaps (bool): Must be set to True if only contiguous motifs (position-specific k-mers) are allowed. By default, no_gaps is False, meaning both gapped and ungapped motifs are searched for.

    - min_precision (float): The minimum precision threshold for keeping a motif. The default value for min_precision is 0.8.

    - min_recall (float): The minimum recall threshold for keeping a motif. The default value for min_precision is 0.
      It is also possible to specify a recall threshold for each motif size. In this case, a dictionary must be specified where
      the motif sizes are keys and the recall values are values. Use the :py:obj:`~immuneML.reports.data_reports.MotifGeneralizationAnalysis` report
      to calibrate the optimal recall threshold given a user-defined precision threshold to ensure generalisability to unseen data.

    - min_true_positives (int): The minimum number of true positive sequences that a motif needs to occur in. The default value for min_true_positives is 10.

    - candidate_motif_filepath (str): Optional filepath for pre-filterd candidate motifs. This may be used to save time. Only the given candidate motifs are considered.
      When this encoder has been run previously, a candidate motifs file named 'all_candidate_motifs.tsv' will have been exported. This file contains all
      possible motifs with high enough min_true_positives without applying precision and recall thresholds.
      The file must be a tab-separated file, structured as follows:

      ========  ==============
      indices    amino_acids
      ========  ==============
      1&2&3      A&G&C
      5&7        E&D
      ========  ==============

      The example above contains two motifs: AGC in positions 123, and E-D in positions 5-7 (with a gap at position 6).

    - label (str): The name of the binary label to train the encoder for. This is only necessary when the dataset contains multiple labels.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_motif_encoder:
                    MotifEncoder:
                        max_positions: 4
                        min_precision: 0.8
                        min_recall:  # different recall thresholds for each motif size
                            1: 0.5   # For shorter motifs, a stricter recall threshold is used
                            2: 0.1
                            3: 0.01
                            4: 0.001
                        min_true_positives: 10

    """

    def __init__(self, max_positions: int = None, min_positions: int = None,
                 min_precision: float = None, min_recall: dict = None,
                 min_true_positives: int = None, no_gaps: bool = False,
                 candidate_motif_filepath: str = None, label: str = None, name: str = None):
        super().__init__(name=name)
        self.max_positions = max_positions
        self.min_positions = min_positions
        self.min_precision = min_precision
        self.min_recall = min_recall
        self.min_true_positives = min_true_positives
        self.no_gaps = no_gaps
        self.candidate_motif_filepath = Path(candidate_motif_filepath) if candidate_motif_filepath is not None else None
        self.learned_motif_filepath = None

        self.label = label
        self.context = None

    @staticmethod
    def _prepare_parameters(max_positions: int = None, min_positions: int = None, no_gaps: bool = None, min_precision: float = None, min_recall: dict = None,
                            min_true_positives: int = None, candidate_motif_filepath: str = None, label: str = None, name: str = None):

        location = MotifEncoder.__name__

        ParameterValidator.assert_type_and_value(max_positions, int, location, "max_positions", min_inclusive=1)
        ParameterValidator.assert_type_and_value(min_positions, int, location, "min_positions", min_inclusive=1)
        ParameterValidator.assert_type_and_value(no_gaps, bool, location, "no_gaps")
        assert max_positions >= min_positions, f"{location}: max_positions ({max_positions}) must be greater than or equal to min_positions ({min_positions})"

        ParameterValidator.assert_type_and_value(min_precision, (int, float), location, "min_precision",
                                                 min_inclusive=0, max_inclusive=1)
        ParameterValidator.assert_type_and_value(min_true_positives, int, location, "min_true_positives",
                                                 min_inclusive=1)

        if isinstance(min_recall, dict):
            assert set(min_recall.keys()) == set(range(min_positions,
                                                       max_positions + 1)), f"{location}: {min_recall} is not a valid value for parameter min_recall. " \
                                                                            f"When setting separate recall cutoffs for each motif size, the keys of the dictionary " \
                                                                            f"must equal to {list(range(min_positions, max_positions + 1))}."
            for recall_cutoff in min_recall.values():
                assert isinstance(recall_cutoff, (int,
                                                  float)) or recall_cutoff is None, f"{location}: {min_recall} is not a valid value for parameter min_recall. " \
                                                                                    f"When setting separate recall cutoffs for each motif size, the values of the dictionary " \
                                                                                    f"must be numeric or None."

            min_recall = {key: value if isinstance(value, (int, float)) else 1 for key, value in min_recall.items()}

        else:
            ParameterValidator.assert_type_and_value(min_recall, (int, float), location, "min_recall", min_inclusive=0,
                                                     max_inclusive=1)
            min_recall = {motif_size: min_recall for motif_size in range(min_positions, max_positions + 1)}

        if candidate_motif_filepath is not None:
            PositionalMotifHelper.check_motif_filepath(candidate_motif_filepath, location, "candidate_motif_filepath")

        if label is not None:
            ParameterValidator.assert_type_and_value(label, str, location, "label")

        return {
            "max_positions": max_positions,
            "min_positions": min_positions,
            "no_gaps": no_gaps,
            "min_precision": min_precision,
            "min_recall": min_recall,
            "min_true_positives": min_true_positives,
            "candidate_motif_filepath": candidate_motif_filepath,
            "label": label,
            "name": name,
        }

    @staticmethod
    def build_object(dataset=None, **params):
        if isinstance(dataset, SequenceDataset):
            prepared_params = MotifEncoder._prepare_parameters(**params)
            return MotifEncoder(**prepared_params)
        else:
            raise ValueError(f"{MotifEncoder.__name__} is not defined for dataset types which are not SequenceDataset.")

    def encode(self, dataset, params: EncoderParams):
        if params.learn_model:
            EncoderHelper.check_positive_class_labels(params.label_config, MotifEncoder.__name__)
            return self._encode_data(dataset, params)
        else:
            learned_motifs = PositionalMotifHelper.read_motifs_from_file(self.learned_motif_filepath)
            return self.get_encoded_dataset_from_motifs(dataset, learned_motifs, params)

    def _encode_data(self, dataset, params: EncoderParams):
        learned_motifs = self._compute_motifs(dataset, params)

        self.learned_motif_filepath = params.result_path / "significant_motifs.tsv"
        self.motif_stats_filepath = params.result_path / "motif_stats.tsv"

        PositionalMotifHelper.write_motifs_to_file(learned_motifs, self.learned_motif_filepath)
        self._write_motif_stats(learned_motifs, self.motif_stats_filepath)

        return self.get_encoded_dataset_from_motifs(dataset, learned_motifs, params)

    def _compute_motifs(self, dataset, params):
        motifs = self._prepare_candidate_motifs(dataset, params)

        y_true = self._get_y_true(dataset, params.label_config)

        motifs = self._filter_motifs(motifs, dataset, y_true, params.pool_size, generalized=False)

        # Option disabled for now
        # if self.generalize_motifs:
        #     motifs += self._filter_motifs(PositionalMotifHelper.get_generalized_motifs(motifs), dataset, y_true, params.pool_size, generalized=True)

        return motifs

    def _write_motif_stats(self, learned_motifs, motif_stats_filepath):
        try:
            data = {}

            data["motif_size"] = list(range(self.min_positions, self.max_positions + 1))
            data["min_precision"] = [self.min_precision] * len(data["motif_size"])
            data["min_recall"] = [self.min_recall.get(motif_size, 1) for motif_size in range(self.min_positions, self.max_positions + 1)]

            all_motif_sizes = [len(motif[0]) for motif in learned_motifs]
            data["n_motifs"] = [all_motif_sizes.count(motif_size) for motif_size in
                                range(self.min_positions, self.max_positions + 1)]

            df = pd.DataFrame(data)
            df.to_csv(motif_stats_filepath, index=False, sep="\t")
        except Exception as e:
            logging.warning(f"{MotifEncoder.__name__}: could not write motif stats. Exception was: {e},"
                          f"{data}")

    def get_encoded_dataset_from_motifs(self, dataset, motifs, params):
        labels = EncoderHelper.encode_element_dataset_labels(dataset, params.label_config)

        examples, feature_names, feature_annotations = self._construct_encoded_data_matrix(dataset, motifs,
                                                                                           params.label_config,
                                                                                           params.pool_size)

        self._export_confusion_matrix(params.result_path, feature_annotations)

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(examples=examples,
                                                   labels=labels,
                                                   feature_names=feature_names,
                                                   feature_annotations=feature_annotations,
                                                   example_ids=dataset.get_example_ids(),
                                                   encoding=MotifEncoder.__name__,
                                                   info={"candidate_motif_filepath": self.candidate_motif_filepath,
                                                         "learned_motif_filepath": self.learned_motif_filepath,
                                                         "positive_class": self._get_positive_class(
                                                             params.label_config),
                                                         "sequence_type": params.sequence_type,
                                                         'region_type': params.region_type})

        return encoded_dataset

    def _export_confusion_matrix(self, result_path, feature_annotations):
        try:
            PathBuilder.build(result_path)
            feature_annotations.to_csv(result_path / "confusion_matrix.tsv", index=False, sep="\t")
        except Exception as e:
            logging.exception(f"MotifEncoder: An exception occurred while exporting the confusion matrix: {e}")

    def _prepare_candidate_motifs(self, dataset, params):
        full_dataset = EncoderHelper.get_current_dataset(dataset, self.context)
        candidate_motifs = self._get_candidate_motifs(full_dataset, params.pool_size)
        assert len(candidate_motifs) > 0, f"{MotifEncoder.__name__}: no candidate motifs were found. " \
                                          f"Please try decreasing the value for parameter 'min_true_positives'."

        self.candidate_motif_filepath = params.result_path / "all_candidate_motifs.tsv"
        PositionalMotifHelper.write_motifs_to_file(candidate_motifs, self.candidate_motif_filepath)

        return candidate_motifs

    def _get_candidate_motifs(self, full_dataset, pool_size=4):
        '''Returns all candidate motifs, which are either read from the input file or computed by finding
        all motifs occuring in at least a given number of sequences of the full dataset.'''
        if self.candidate_motif_filepath is None:
            return CacheHandler.memo_by_params(self._build_candidate_motifs_params(full_dataset),
                                               lambda: self._compute_candidate_motifs(full_dataset, pool_size))
        else:
            return PositionalMotifHelper.read_motifs_from_file(self.candidate_motif_filepath)

    def _build_candidate_motifs_params(self, dataset: SequenceDataset):
        return (("dataset_identifier", dataset.identifier),
                ("sequence_ids", tuple(dataset.get_example_ids()),
                 ("max_positions", self.max_positions),
                 ("min_positions", self.min_positions),
                 ("min_true_positives", self.min_true_positives)))

    def _compute_candidate_motifs(self, full_dataset, pool_size=4):
        np_sequences = PositionalMotifHelper.get_numpy_sequence_representation(full_dataset)
        params = PositionalMotifParams(max_positions=self.max_positions, min_positions=self.min_positions,
                                       count_threshold=self.min_true_positives, no_gaps=self.no_gaps,
                                       pool_size=pool_size)
        return PositionalMotifHelper.compute_all_candidate_motifs(np_sequences, params)

    def _get_y_true(self, dataset, label_config: LabelConfiguration):
        labels = EncoderHelper.encode_element_dataset_labels(dataset, label_config)

        label_name = self._get_label_name(label_config)
        label = label_config.get_label_object(label_name)

        return np.array([cls == label.positive_class for cls in labels[label_name]])

    def _get_positive_class(self, label_config):
        label_name = self._get_label_name(label_config)
        label = label_config.get_label_object(label_name)

        return label.positive_class

    def _get_label_name(self, label_config: LabelConfiguration):
        if self.label is not None:
            assert self.label in label_config.get_labels_by_name(), f"{MotifEncoder.__name__}: specified label " \
                                                                    f"'{self.label}' was not present among the dataset labels: " \
                                                                    f"{', '.join(label_config.get_labels_by_name())}"
            label_name = self.label
        else:
            label_name = EncoderHelper.get_single_label_name_from_config(label_config, MotifEncoder.__name__)

        return label_name

    def check_filtered_motifs(self, filtered_motifs):
        assert len(filtered_motifs) > 0, f"{MotifEncoder.__name__}: no significant motifs were found. " \
                                         f"Please try decreasing the values for parameters 'min_precision' or 'min_recall'"

    def _get_recall_repr(self):
        '''Returns a string representation of the recall cutoff.'''
        if len(set(self.min_recall.values())) == 1:
            return str(list(self.min_recall.values())[0])
        else:
            return ", ".join([f"{recall} (motif size {motif_size})" for motif_size, recall in self.min_recall.items()])

    def _filter_motifs(self, candidate_motifs, dataset, y_true, pool_size, generalized=False):
        motif_type = "generalized motifs" if generalized else "motifs"

        logging.info(
            f"{MotifEncoder.__name__}: filtering {len(candidate_motifs)} {motif_type} with precision >= {self.min_precision} and recall >= {self._get_recall_repr()}")

        np_sequences = PositionalMotifHelper.get_numpy_sequence_representation(dataset)
        weights = None

        with Pool(pool_size) as pool:
            partial_func = partial(PositionalMotifHelper.check_motif, np_sequences=np_sequences, y_true=y_true, weights=weights,
                                   min_precision=self.min_precision, min_true_positives=self.min_true_positives, min_recall=self.min_recall)

            filtered_motifs = list(filter(None, pool.map(partial_func, candidate_motifs)))

        if not generalized:
            self.check_filtered_motifs(filtered_motifs)

        logging.info(f"{MotifEncoder.__name__}: filtering {motif_type} done, {len(filtered_motifs)} motifs left")

        return filtered_motifs

    def _construct_encoded_data_matrix(self, dataset, motifs, label_config, number_of_processes):
        feature_names = [PositionalMotifHelper.motif_to_string(indices, amino_acids, motif_sep="-", newline=False)
                         for indices, amino_acids in motifs]

        weights = None # TODO: have a better way to handle weights
        y_true = self._get_y_true(dataset, label_config)
        np_sequences = PositionalMotifHelper.get_numpy_sequence_representation(dataset)

        logging.info(f"{MotifEncoder.__name__}: building encoded data matrix...")

        with Pool(number_of_processes) as pool:
            predictions = pool.starmap(partial(MotifEncoder._test_motif, np_sequences=np_sequences), motifs)
            conf_matrix_raw = np.array(pool.map(partial(MotifEncoder._get_confusion_matrix, y_true=y_true, weights=None), predictions))

            if weights is not None:
                conf_matrix_weighted = np.array(pool.map(partial(MotifEncoder._get_confusion_matrix, y_true=y_true, weights=weights), predictions))
            else:
                conf_matrix_weighted = None

            # precision_scores = pool.map(partial(self._get_precision, y_true=y_true, weights=weights), predictions)
            # recall_scores = pool.map(partial(self._get_recall, y_true=y_true, weights=weights), predictions)
            # tp_counts = pool.map(partial(self._get_tp, y_true=y_true), predictions)

        logging.info(f"{MotifEncoder.__name__}: building encoded data matrix done")

        prefix = "weighted_" if weights is not None else ""

        feature_annotations = self._get_feature_annotations(feature_names, conf_matrix_raw, conf_matrix_weighted)

        return np.column_stack(predictions), feature_names, feature_annotations

    def _get_feature_annotations(self, feature_names, conf_matrix_raw, conf_matrix_weighted):
        feature_annotations_mapping = {"feature_names": feature_names,
                                       "TN": conf_matrix_raw.T[0],
                                       "FP": conf_matrix_raw.T[1],
                                       "FN": conf_matrix_raw.T[2],
                                       "TP": conf_matrix_raw.T[3]}

        if conf_matrix_weighted is not None:
            feature_annotations_mapping["weighted_TN"] = conf_matrix_weighted.T[0]
            feature_annotations_mapping["weighted_FP"] = conf_matrix_weighted.T[1]
            feature_annotations_mapping["weighted_FN"] = conf_matrix_weighted.T[2]
            feature_annotations_mapping["weighted_TP"] = conf_matrix_weighted.T[3]

        return pd.DataFrame(feature_annotations_mapping)

    def _get_predictions(self, np_sequences, motifs, number_of_processes):
        with Pool(number_of_processes) as pool:
            partial_func = partial(MotifEncoder._test_motif, np_sequences=np_sequences)
            predictions = pool.starmap(partial_func, motifs)

        return predictions

    @staticmethod
    def _test_motif(indices, amino_acids, np_sequences):
        return PositionalMotifHelper.test_motif(np_sequences=np_sequences, indices=indices, amino_acids=amino_acids)

    @staticmethod
    def _get_confusion_matrix(pred, y_true, weights):
        return confusion_matrix(y_true=y_true, y_pred=pred, sample_weight=weights).ravel()

    def set_context(self, context: dict):
        self.context = context
        return self



import copy

import pandas as pd
from scipy.sparse import issparse

from immuneML.data_model.datasets.Dataset import Dataset


class DataReshaper:

    @staticmethod
    def reshape(dataset: Dataset, labels=None):
        """
        Takes a 2D matrix of values from the encoded data and reshapes it to long format,
        retaining the column and row annotations. This is for ease of use in plotting the data.
        It is suggested that some sort of filtering is done first, otherwise the memory usage may explode, as
        the resulting data frame is of shape
        (matrix.shape[0] * matrix.shape[1], labels.shape[0] + feature_annotations.shape[1] + 1)
        """
        if labels is None:
            row_annotations = pd.DataFrame(dataset.encoded_data.labels)
        else:
            row_annotations = pd.DataFrame(dataset.get_metadata(labels, return_df=True))

        row_annotations["example_id"] = dataset.encoded_data.example_ids

        if dataset.encoded_data.feature_annotations is not None:
            column_annotations = copy.deepcopy(dataset.encoded_data.feature_annotations)
            column_annotations["feature"] = dataset.encoded_data.feature_names
        else:
            column_annotations = pd.DataFrame({'feature': dataset.encoded_data.feature_names})

        matrix = dataset.encoded_data.examples
        matrix_1d = matrix.toarray().ravel() if issparse(matrix) else matrix.ravel()

        column_annotations = pd.concat([column_annotations]*matrix.shape[0], ignore_index=True)
        row_annotations = pd.DataFrame(row_annotations.values.repeat(matrix.shape[1], axis=0), columns=row_annotations.columns)
        data = pd.concat([row_annotations.reset_index(drop=True), column_annotations.reset_index(drop=True), pd.DataFrame({"value": matrix_1d})], axis=1)

        for column in data.columns:
            data[column] = pd.to_numeric(data[column], errors="ignore")

        return data



from pathlib import Path

from immuneML.data_model.SequenceParams import Chain
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.preprocessing.filters.Filter import Filter
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class ChainRepertoireFilter(Filter):
    """
    This filter has two options: it can remove repertoires from the dataset which have any chain other than the
    specified one (e.g., keep  only TRB) or it can remove the sequences from the repertoire which do not have the
    desired chain.

    Since the filter may remove repertoires/sequences from the dataset (examples in machine learning setting), it
    cannot be used with :ref:`TrainMLModel` instruction. If you want to filter out repertoires including a given chain,
    see :ref:`DatasetExport` instruction with preprocessing.

    **Dataset types:**

    - RepertoireDataset

    **Specification arguments:**

    - keep_chains (list): Which chains should be kept, valid values are "TRA", "TRB", "IGH", "IGL", "IGK"

    - remove_only_sequences (bool): Whether to remove only sequences with different chain than "keep_chain" (true) in
      case of repertoire datasets; default is false

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        preprocessing_sequences:
            my_preprocessing:
                - my_filter:
                    ChainRepertoireFilter:
                        keep_chains: [TRB]
                        remove_only_sequences: true

    """

    def __init__(self, keep_chains: list, remove_only_sequences: bool = False, result_path: Path = None):
        super().__init__(result_path)
        ParameterValidator.assert_type_and_value(keep_chains, list, "ChainRepertoireFilter", "keep_chains")
        self.keep_chains = [Chain.get_chain(keep_chain) for keep_chain in keep_chains]
        self.remove_only_sequences = remove_only_sequences

    def process_dataset(self, dataset: RepertoireDataset, result_path: Path, number_of_processes=1):
        self.check_dataset_type(dataset, [RepertoireDataset], "ChainRepertoireFilter")
        processed_dataset = dataset.clone()
        self.result_path = PathBuilder.build(result_path)
        return self._filter_repertoire_dataset(processed_dataset)

    def _filter_repertoire_dataset(self, processed_dataset: RepertoireDataset):

        repertoires = []
        indices = []
        valid_chains = [chain.value for chain in self.keep_chains]
        for index, repertoire in enumerate(processed_dataset.get_data()):

            if self.remove_only_sequences:
                data = repertoire.data
                data = data[[Chain.get_chain(l).value in valid_chains for l in data.locus.tolist()]]
                if len(data) == 0:
                    continue

                new_repertoire = Repertoire.build_from_dc_object(self.result_path, repertoire.metadata, data=data,
                                                                 filename_base=repertoire.data_filename.stem)
                repertoires.append(new_repertoire)
                indices.append(index)
            else:
                if all(Chain.get_chain(l).value in valid_chains for l in repertoire.data.locus.tolist()):
                    repertoires.append(repertoire)
                    indices.append(index)

        processed_dataset.repertoires = repertoires
        processed_dataset.metadata_file = self._build_new_metadata(processed_dataset, indices)

        self.check_dataset_not_empty(processed_dataset, "ChainRepertoireFilter")

        return processed_dataset

    def keeps_example_count(self) -> bool:
        return False



import hashlib
from multiprocessing.pool import Pool

import math
from typing import Union

import dill
import numpy as np

from immuneML.analysis.entropy_calculations.EntropyCalculator import EntropyCalculator
from immuneML.caching.CacheHandler import CacheHandler
from immuneML.caching.CacheObjectType import CacheObjectType
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.diversity_encoding.EvennessProfileEncoder import EvennessProfileEncoder
from immuneML.util.Logger import log


class EvennessProfileRepertoireEncoder(EvennessProfileEncoder):

    def _encode_new_dataset(self, dataset, params: EncoderParams):

        encoded_data = self._encode_data(dataset, params)

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = encoded_data

        return encoded_dataset

    def _encode_examples(self, dataset, params: EncoderParams):

        arguments = [(repertoire.identifier, dill.dumps(repertoire), params) for repertoire in dataset.repertoires]

        with Pool(params.pool_size) as pool:
            chunksize = math.floor(dataset.get_example_count()/params.pool_size) + 1
            repertoires = pool.starmap(self.get_encoded_repertoire, arguments, chunksize=chunksize)

        encoded_repertoire_list, repertoire_names, labels = zip(*repertoires)

        encoded_labels = {k: [dic[k] for dic in labels] for k in labels[0]} if params.encode_labels else None

        return list(encoded_repertoire_list), list(repertoire_names), encoded_labels

    def get_encoded_repertoire(self, repertoire_id: str, repertoire: Union[bytes, Repertoire], params: EncoderParams):

        params.model = vars(self)
        serialized_repertoire = dill.dumps(repertoire) if isinstance(repertoire, Repertoire) else repertoire

        return CacheHandler.memo_by_params((("encoding_model", params.model),
                                            ("labels", params.label_config.get_labels_by_name()),
                                            ("repertoire_id", repertoire_id)),
                                           lambda: self.encode_repertoire(serialized_repertoire, params),
                                           CacheObjectType.ENCODING_STEP)

    def encode_repertoire(self, repertoire, params: EncoderParams):
        if isinstance(repertoire, bytes):
            repertoire = dill.loads(repertoire)

        alphas = np.linspace(start=params.model["min_alpha"], stop=params.model["max_alpha"], num=params.model["dimension"])

        data = repertoire.data
        counts = data.duplicate_count[np.array(data.vj_in_frame == 'T').flatten()]

        freqs = counts[np.nonzero(counts)]

        evenness_profile = np.array([np.exp(EntropyCalculator.renyi_entropy(freqs, alpha))/len(freqs) for alpha in alphas])

        if params.encode_labels:
            label_config = params.label_config
            labels = dict()
            for label_name in label_config.get_labels_by_name():
                label = repertoire.metadata[label_name]
                labels[label_name] = label
        else:
            labels = None

        return evenness_profile, repertoire.identifier, labels



from collections import Counter

import numpy as np
import pandas as pd

from immuneML.encodings.atchley_kmer_encoding.RelativeAbundanceType import RelativeAbundanceType
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.KmerHelper import KmerHelper


class Util:
    ATCHLEY_FACTOR_COUNT = 5
    ATCHLEY_FACTORS = None

    @staticmethod
    def compute_abundance(sequences: np.ndarray, counts: np.ndarray, k: int, abundance: RelativeAbundanceType):
        func = getattr(Util, f"compute_{abundance.value}")
        return func(sequences, counts, k)

    @staticmethod
    def compute_relative_abundance(sequences: np.ndarray, counts: np.ndarray, k: int) -> dict:
        """
        Computes the relative abundance of k-mers in the repertoire per following equations where C is the template count, T is the total count and
        RA is relative abundance (the output of the function for each k-mer separately):

        .. math::

            C^{kmer}=\\sum_{\\underset{with kmer}{TCR \\beta}} C^{TCR \\beta}

            T^{kmer} = \\sum_{kmer} C^{kmer}

            RA = \\frac{C^{kmer}}{T^{kmer}}

        For more details, please see the original publication: Ostmeyer J, Christley S, Toby IT, Cowell LG. Biophysicochemical motifs in T cell
        receptor sequences distinguish repertoires from tumor-infiltrating lymphocytes and adjacent healthy tissue. Cancer Res. Published online
        January 1, 2019:canres.2292.2018. `doi:10.1158/0008-5472.CAN-18-2292 <https://cancerres.aacrjournals.org/content/canres/79/7/1671.full.pdf>`_

        Arguments:

            sequences: an array of (amino acid) sequences (corresponding to a repertoire)
            counts: an array of counts for each of the sequences
            k: the length of the k-mer (in the publication referenced above, k is 4)

        Returns:

            a dictionary where keys are k-mers and values are their relative abundances in the given list of sequences

        """

        c_kmers = Counter()
        for index, sequence in enumerate(sequences):
            kmers = KmerHelper.create_kmers_from_string(sequence, k)
            c_kmers += {kmer: counts[index] for kmer in kmers}

        t_kmers = sum(c_kmers.values())

        return {kmer: c_kmers[kmer] / t_kmers for kmer in c_kmers.keys()}

    @staticmethod
    def compute_tcrb_relative_abundance(sequences: np.ndarray, counts: np.ndarray, k: int) -> dict:
        """
        Computes the relative abundance of k-mers in the repertoire per following equations where C is the template count for the given receptor
        sequence, T is the total count across all receptor sequences. The relative abundance per receptor sequence is then computed and only the
        maximum sequence abudance was used for the k-mer so that the k-mer's relative abundance is equal to the abundance of the most frequent
        receptor sequence in which the receptor appears:

        .. math::

            T^{TCR \\beta} = \\sum_{TCR\\beta} C^{TCR\\beta}

            RA^{TCR\\beta} = \\frac{C^{TCR\\beta}}{T^{TCR\\beta}}

            RA = \\max_{\\underset{with \\, kmer}{TCR\\beta}} {RA^{TCR \\beta}}

        For more details, please see the original publication: Ostmeyer J, Christley S, Toby IT, Cowell LG. Biophysicochemical motifs in T cell
        receptor sequences distinguish repertoires from tumor-infiltrating lymphocytes and adjacent healthy tissue. Cancer Res. Published online
        January 1, 2019:canres.2292.2018. `doi:10.1158/0008-5472.CAN-18-2292 <https://cancerres.aacrjournals.org/content/canres/79/7/1671.full.pdf>`_

        Arguments:

            sequences: an array of (amino acid) sequences (corresponding to a repertoire)
            counts: an array of counts for each of the sequences
            k: the length of the k-mer (in the publication referenced above, k is 4)

        Returns:

            a dictionary where keys are k-mers and values are their relative abundances in the given list of sequences

        """
        relative_abundance = {}
        total_count = np.sum(counts)
        relative_abundance_per_sequence = counts / total_count
        for index, sequence in enumerate(sequences):
            kmers = KmerHelper.create_kmers_from_string(sequence, k)
            for kmer in kmers:
                if kmer not in relative_abundance or relative_abundance[kmer] < relative_abundance_per_sequence[index]:
                    relative_abundance[kmer] = relative_abundance_per_sequence[index]

        return relative_abundance

    @staticmethod
    def get_atchely_factors(kmers: list, k: int) -> pd.DataFrame:
        """
        Returns values of Atchley factors for each amino acid in the sequence. The data was downloaded from the publication:
        Atchley WR, Zhao J, Fernandes AD, Drüke T. Solving the protein sequence metric problem. PNAS. 2005;102(18):6395-6400.
        `doi:10.1073/pnas.0408677102 <https://www.pnas.org/content/102/18/6395>`_

        Arguments:

            kmers: a list of amino acid sequences
            k: length of k-mers

        Returns:

            values of Atchley factors for each amino acid in the sequence

        """
        if Util.ATCHLEY_FACTORS is None:
            Util.ATCHLEY_FACTORS = pd.read_csv(EnvironmentSettings.root_path / "immuneML/encodings/atchley_kmer_encoding/atchley_factors.csv",
                                               index_col='amino_acid')

        factors = np.zeros((len(kmers), Util.ATCHLEY_FACTOR_COUNT * k))
        for index, kmer in enumerate(kmers):
            factors[index] = np.concatenate([Util.ATCHLEY_FACTORS.loc[amino_acid].values for amino_acid in kmer])

        factors_df = pd.DataFrame(factors, index=kmers)

        return factors_df



from pathlib import Path

from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.hyperparameter_optimization.states.HPLabelState import HPLabelState


class HPAssessmentState:

    def __init__(self, split_index: int, train_val_dataset, test_dataset, path: Path, label_configuration: LabelConfiguration):
        self.split_index = split_index
        self.train_val_dataset = train_val_dataset
        self.test_dataset = test_dataset
        self.path = path
        self.train_val_data_reports = []
        self.test_data_reports = []

        # computed
        self.label_states = {label_name: HPLabelState(label_name, label_configuration.get_auxiliary_labels(label_name))
                             for label_name in label_configuration.get_labels_by_name()}



import logging
from typing import Tuple, List

import pandas as pd

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.SequenceParams import RegionType, Chain
from immuneML.data_model.datasets.ElementDataset import ElementDataset


class TCRdistHelper:

    @staticmethod
    def get_chains(dataset: ElementDataset) -> list:
        """
        Returns the chains used in the dataset.

        Args:
            dataset: receptor dataset for which all pairwise distances between receptors will be computed

        Returns:
            a list of chains used in the dataset
        """
        return [str(Chain.get_chain(el)).lower() for el in set(dataset.data.locus.tolist())]

    @staticmethod
    def compute_tcr_dist(dataset: ElementDataset, label_names: list, cores: int = 1, cdr3_only: bool = False):
        return CacheHandler.memo_by_params((('dataset_identifier', dataset.identifier), ("type", "TCRrep")),
                                           lambda: TCRdistHelper._compute_tcr_dist(dataset, label_names, cores, cdr3_only))

    @staticmethod
    def _compute_tcr_dist(dataset: ElementDataset, label_names: list, cores: int, cdr3_only: bool = False):
        """
        Computes the tcrdist distances by creating a TCRrep object and calling compute_distances() function.

        Parameters `ntrim` and `ctrim` in TCRrep object for CDR3 are adjusted to account for working with IMGT CDR3 definition if IMGT CDR3 was set
        as region_type for the dataset upon importing. `deduplicate` parameter is set to False as we assume that we work with clones in immuneML,
        and not individual receptors.

        Args:
            dataset: receptor dataset for which all pairwise distances between receptors will be computed
            label_names: a list of label names (e.g., specific epitopes) to be used for later classification or reports
            cores: how many cpus to use for computation

        Returns:
            an instance of TCRrep object with computed pairwise distances between all receptors in the dataset

        """
        from tcrdist.repertoire import TCRrep

        df, chains = TCRdistHelper.prepare_tcr_dist_dataframe(dataset, label_names)
        tcr_rep = TCRrep(cell_df=df, chains=chains, organism=dataset.labels["organism"], cpus=cores,
                         deduplicate=False, compute_distances=False)

        if cdr3_only:
            for chain in chains:
                for attr_prefix in ['metrics', 'kargs', 'weights']:
                    setattr(tcr_rep, f'{attr_prefix}_{chain[0]}',
                            {key: value for key, value in getattr(tcr_rep, f'{attr_prefix}_{chain[0]}').items() if 'cdr3' in key})

        tcr_rep.compute_distances()

        return tcr_rep

    @staticmethod
    def add_default_allele_to_v_gene(v_gene: str):
        if v_gene is not None and "*" not in v_gene:
            return f"{v_gene}*01"
        else:
            return v_gene

    @staticmethod
    def prepare_tcr_dist_dataframe(dataset: ElementDataset, label_names: list) -> Tuple[pd.DataFrame, List[str]]:

        df = dataset.data.topandas()

        if "subject" not in df:
            df['subject'] = "sub" + df['cell_id']

        df.loc[df['v_call'].str.contains("\*"), 'v_call'] = [TCRdistHelper.add_default_allele_to_v_gene(el) for el in
                                                             df.loc[df['v_call'].str.contains("\*"), 'v_call']]
        df.loc[df['j_call'].str.contains("\*"), 'j_call'] = [TCRdistHelper.add_default_allele_to_v_gene(el) for el in
                                                             df.loc[df['j_call'].str.contains("\*"), 'j_call']]
        unique_chains = [str(Chain.get_chain(el)).lower() for el in df['locus'].unique().tolist()]

        df['clone_id'] = df['cell_id' if len(unique_chains) == 2 else 'sequence_id']
        cols_to_keep = ['cdr3', 'cdr3_aa', 'v_call', 'j_call', 'duplicate_count', 'subject', 'clone_id'] + label_names

        df_alpha, df_beta = None, None

        if 'alpha' in unique_chains:

            df_alpha = (df[df.locus == 'TRA'][cols_to_keep]
                        .rename(columns={"cdr3_aa": "cdr3_a_aa", "cdr3": "cdr3_a_nucseq", "v_call": "v_a_gene",
                                         'j_call': "j_a_gene", "duplicate_count": "count"}))
            df_alpha.loc[:, 'count'] = [1 if el in [-1, None] else el for el in df_alpha['count']]
            if len(unique_chains) == 1:
                df = df_alpha

        if 'beta' in unique_chains:

            df_beta = (df[df.locus == 'TRB'][cols_to_keep].rename(
                columns={"cdr3_aa": "cdr3_b_aa", "cdr3": "cdr3_b_nucseq", "v_call": "v_b_gene",
                         'j_call': "j_b_gene", "duplicate_count": "count"}))
            df_beta.loc[:, 'count'] = [1 if el in [-1, None] else el for el in df_beta['count']]
            if len(unique_chains) == 1:
                df = df_beta

        if len(unique_chains) == 2:
            df = df_alpha.merge(df_beta, on=['clone_id', 'subject', 'count'] + label_names)

        return df, unique_chains



from pathlib import Path

from immuneML.hyperparameter_optimization.strategy.HPOptimizationStrategy import HPOptimizationStrategy


class HPSelectionState:

    def __init__(self, train_datasets, val_datasets, path: Path, hp_strategy: HPOptimizationStrategy):
        self.train_datasets = train_datasets
        self.val_datasets = val_datasets
        self.path = path
        self.hp_strategy = hp_strategy.clone()
        self.hp_items = {str(hp_setting): [] for hp_setting in self.hp_strategy.hp_settings}
        self.train_data_reports = []
        self.val_data_reports = []
        self.data_reports = []

    @property
    def optimal_hp_setting(self):
        return self.hp_strategy.get_optimal_hps()



import copy

from sklearn.ensemble import GradientBoostingClassifier

from immuneML.ml_methods.classifiers.SklearnMethod import SklearnMethod
from scripts.specification_util import update_docs_per_mapping


class GradientBoosting(SklearnMethod):
    """
    Gradient Boosting classifier which wraps scikit-learn's GradientBoostingClassifier.
    Input arguments for the method are the same as supported by scikit-learn (see `GradientBoostingClassifier scikit-learn documentation
    <https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.GradientBoostingClassifier.html>`_ for details).

    For usage instructions, check :py:obj:`~immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod`.

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_gradient_boosting:
                    GradientBoosting:
                        # arguments as defined by scikit-learn
                        n_estimators: 100
                        learning_rate: 0.1
                        max_depth: 3
                        random_state: 42

        """

    def __init__(self, parameter_grid: dict = None, parameters: dict = None):
        super(GradientBoosting, self).__init__(parameter_grid=parameter_grid, parameters=parameters)

    def _get_ml_model(self, cores_for_training: int = 2, X=None):
        params = copy.deepcopy(self._parameters) if self._parameters is not None else {}
        return GradientBoostingClassifier(**params)

    def can_predict_proba(self) -> bool:
        return True

    def can_fit_with_example_weights(self) -> bool:
        return True

    def get_params(self, for_refitting=False):
        params = copy.deepcopy(self.model.get_params())
        if not for_refitting:
            params["feature_importances"] = self.model.feature_importances_.tolist()
        return params

    @staticmethod
    def get_documentation():
        doc = str(GradientBoosting.__doc__)

        mapping = {
            "For usage instructions, check :py:obj:`~immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod`.": SklearnMethod.get_usage_documentation(
                "GradientBoosting"),
        }

        doc = update_docs_per_mapping(doc, mapping)
        return doc


from immuneML.IO.dataset_export.DataExporter import DataExporter
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReflectionHandler import ReflectionHandler


def parse_exporters(instruction, location):
    if instruction["export_formats"] is not None:
        class_path = "dataset_export/"
        ParameterValidator.assert_all_in_valid_list(instruction["export_formats"],
                                                    ReflectionHandler.all_nonabstract_subclass_basic_names(DataExporter, 'Exporter', class_path),
                                                    location=location, parameter_name="export_formats")
        exporters = [ReflectionHandler.get_class_by_name(f"{item}Exporter", class_path) for item in instruction["export_formats"]]
    else:
        exporters = None

    return exporters



import abc
import math

import numpy as np

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model import bnp_util
from immuneML.data_model.AIRRSequenceSet import AIRRSequenceSet
from immuneML.data_model.SequenceParams import RegionType
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.EncoderHelper import EncoderHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReflectionHandler import ReflectionHandler


class OneHotEncoder(DatasetEncoder):
    """
    One-hot encoding for repertoires, sequences or receptors. In one-hot encoding, each alphabet character
    (amino acid or nucleotide) is replaced by a sparse vector with one 1 and the rest zeroes. The position of the
    1 represents the alphabet character.

    **Dataset type:**

    - SequenceDatasets

    - ReceptorDatasets

    - RepertoireDatasets


    **Specification arguments:**

    - use_positional_info (bool): whether to include features representing the positional information.
      If True, three additional feature vectors will be added, representing the sequence start, sequence middle
      and sequence end. The values in these features are scaled between 0 and 1. A graphical representation of
      the values of these vectors is given below.

    .. code-block:: console

          Value of sequence start:         Value of sequence middle:        Value of sequence end:

        1 \                              1    /‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾\         1                          /
           \                                 /                   \                                  /
            \                               /                     \                                /
        0    \_____________________      0 /                       \      0  _____________________/
          <----sequence length---->        <----sequence length---->         <----sequence length---->


    - distance_to_seq_middle (int): only applies when use_positional_info is True. This is the distance from the edge
      of the CDR3 sequence (IMGT positions 105 and 117) to the portion of the sequence that is considered 'middle'.
      For example: if distance_to_seq_middle is 6 (default), all IMGT positions in the interval [111, 112)
      receive positional value 1.
      When using nucleotide sequences: note that the distance is measured in (amino acid) IMGT positions.
      If the complete sequence length is smaller than 2 * distance_to_seq_middle, the maximum value of the
      'start' and 'end' vectors will not reach 0, and the maximum value of the 'middle' vector will not reach 1.
      A graphical representation of the positional vectors with a too short sequence is given below:


    .. code-block:: console

        Value of sequence start         Value of sequence middle        Value of sequence end:
        with very short sequence:       with very short sequence:       with very short sequence:

             1 \                               1                                 1    /
                \                                                                    /
                 \                                /\                                /
             0                                 0 /  \                            0
               <->                               <-->                               <->

    - flatten (bool): whether to flatten the final onehot matrix to a 2-dimensional matrix [examples, other_dims_combined]
      This must be set to True when using onehot encoding in combination with scikit-learn ML methods (inheriting :py:obj:`~source.ml_methods.SklearnMethod.SklearnMethod`),
      such as :ref:`LogisticRegression`, :ref:`SVM`, :ref:`SVC`, :ref:`RandomForestClassifier` and :ref:`KNN`.

    - sequence_type: whether to use nucleotide or amino acid sequence for encoding. Valid values are 'nucleotide' and 'amino_acid'.

    - region_type: which part of the sequence to encode; e.g., imgt_cdr3, imgt_junction


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                one_hot_vanilla:
                    OneHot:
                        use_positional_info: False
                        flatten: False
                        sequence_type: amino_acid
                        region_type: imgt_cdr3

                one_hot_positional:
                    OneHot:
                        use_positional_info: True
                        distance_to_seq_middle: 3
                        flatten: False
                        sequence_type: nucleotide

    """

    dataset_mapping = {
        "RepertoireDataset": "OneHotRepertoireEncoder",
        "SequenceDataset": "OneHotSequenceEncoder",
        "ReceptorDataset": "OneHotReceptorEncoder"
    }

    def __init__(self, use_positional_info: bool, distance_to_seq_middle: int, flatten: bool, name: str = None,
                 sequence_type: SequenceType = None, region_type: RegionType = None):
        super().__init__(name=name)
        self.use_positional_info = use_positional_info
        self.distance_to_seq_middle = distance_to_seq_middle
        self.flatten = flatten
        self.sequence_type = sequence_type
        self.region_type = region_type
        self.alphabet = EnvironmentSettings.get_sequence_alphabet(self.sequence_type)

        if distance_to_seq_middle:
            self.pos_increasing = [1 / self.distance_to_seq_middle * i for i in range(self.distance_to_seq_middle)]
            self.pos_decreasing = self.pos_increasing[::-1]
        else:
            self.pos_decreasing = None

        if self.sequence_type == SequenceType.NUCLEOTIDE and self.distance_to_seq_middle is not None:  # todo check this / explain in docs
            self.distance_to_seq_middle = self.distance_to_seq_middle * 3

        if self.use_positional_info:
            self.onehot_dimensions = self.alphabet + ["start", "mid", "end"]
        else:
            self.onehot_dimensions = self.alphabet

    @staticmethod
    def _prepare_parameters(use_positional_info: bool, distance_to_seq_middle: int, flatten: bool, sequence_type: str,
                            name: str = None, region_type: str = None):

        location = OneHotEncoder.__name__

        ParameterValidator.assert_type_and_value(use_positional_info, bool, location, "use_positional_info")
        if use_positional_info:
            ParameterValidator.assert_type_and_value(distance_to_seq_middle, int, location, "distance_to_seq_middle",
                                                     min_inclusive=1)
        else:
            distance_to_seq_middle = None

        ParameterValidator.assert_type_and_value(flatten, bool, location, "flatten")
        ParameterValidator.assert_type_and_value(sequence_type, str, location, 'sequence_type')
        ParameterValidator.assert_in_valid_list(sequence_type.upper(), [item.name for item in SequenceType], location,
                                                'sequence_type')
        ParameterValidator.assert_region_type({'region_type': region_type}, location)

        return {"use_positional_info": use_positional_info, "region_type": RegionType[region_type.upper()],
                "distance_to_seq_middle": distance_to_seq_middle,
                "flatten": flatten, "sequence_type": SequenceType[sequence_type.upper()],
                "name": name}

    @staticmethod
    def build_object(dataset=None, **params):
        EncoderHelper.check_dataset_type_available_in_mapping(dataset, OneHotEncoder)

        prepared_params = OneHotEncoder._prepare_parameters(**params)
        encoder = ReflectionHandler.get_class_by_name(OneHotEncoder.dataset_mapping[dataset.__class__.__name__],
                                                      "onehot/")(**prepared_params)
        return encoder

    def encode(self, dataset, params: EncoderParams):
        encoded_dataset = CacheHandler.memo_by_params(self._prepare_caching_params(dataset, params),
                                                      lambda: self._encode_new_dataset(dataset, params))

        return encoded_dataset

    def _prepare_caching_params(self, dataset, params: EncoderParams):
        return (("dataset_identifier", dataset.identifier),
                ("example_identifiers", tuple(dataset.get_example_ids())),
                ("dataset_type", dataset.__class__.__name__),
                ("labels", tuple(params.label_config.get_labels_by_name())),
                ("encoding", OneHotEncoder.__name__),
                ("encoding_params", tuple(vars(self).items())))

    @abc.abstractmethod
    def _encode_new_dataset(self, dataset, params: EncoderParams):
        pass

    def _get_seq_field_name(self, params: EncoderParams) -> str:
        region_type = self.region_type if self.region_type else params.region_type
        sequence_type = self.sequence_type if self.sequence_type else params.sequence_type
        return bnp_util.get_sequence_field_name(region_type, sequence_type)

    def _encode_sequence_list(self, sequences: AIRRSequenceSet, params: EncoderParams, pad_n_sequences: int = None,
                              pad_sequence_len: int = None):
        # Get sequence field based on sequence_type
        sequence_field = self._get_seq_field_name(params)

        # Extract sequences from AIRRSequenceSet
        sequence_array = getattr(sequences, sequence_field)

        sequence_alphabet = "".join(EnvironmentSettings.get_sequence_alphabet(params.sequence_type))

        # noinspection PyTypeChecker
        encoded_sequences = np.array([
            np.pad(np.array((sequence_array[i][..., np.newaxis] == sequence_alphabet).tolist()),
                   [[0, pad_sequence_len - sequence_array.lengths[i]], [0, 0]], mode='constant',
                   constant_values=False) for i in range(len(sequence_array))
        ])

        if encoded_sequences.shape[0] != pad_n_sequences:
            encoded_sequences = np.concatenate([encoded_sequences,
                                                np.zeros(
                                                    (pad_n_sequences - encoded_sequences.shape[0], pad_sequence_len,
                                                     len(sequence_alphabet)))], axis=0)

        # Add positional encoding if needed
        if self.use_positional_info:
            pos_info = ([self._get_imgt_position_weights(seq_len,
                                                         pad_length=pad_sequence_len).T
                         for seq_len in sequence_array.lengths]
                        + [[[0, 0, 0] for _ in range(pad_sequence_len)] for s in range(pad_n_sequences - len(sequence_array))])
            pos_info = np.stack(pos_info)

            # Combine one-hot encoding with positional information
            encoded_sequences = np.concatenate([encoded_sequences, pos_info], axis=-1)

        return encoded_sequences

    def _get_imgt_position_weights(self, seq_length, pad_length=None):
        start_weights = self._get_imgt_start_weights(seq_length)
        mid_weights = self._get_imgt_mid_weights(seq_length)
        end_weights = start_weights[::-1]

        weights = np.array([start_weights, mid_weights, end_weights])

        if pad_length is not None:
            weights = np.pad(weights, pad_width=((0, 0), (0, pad_length - seq_length)))

        return weights

    def _get_imgt_mid_weights(self, seq_length):
        mid_len = seq_length - (self.distance_to_seq_middle * 2)

        if mid_len >= 0:
            mid_weights = self.pos_increasing + [1] * mid_len + self.pos_decreasing
        else:
            left_idx = math.ceil(seq_length / 2)
            right_idx = math.floor(seq_length / 2)

            mid_weights = self.pos_increasing[:left_idx] + self.pos_decreasing[-right_idx:]

        return mid_weights

    def _get_imgt_start_weights(self, seq_length):
        diff = (seq_length - self.distance_to_seq_middle) - 1
        if diff >= 0:
            start_weights = [1] + self.pos_decreasing + [0] * diff
        else:
            start_weights = [1] + self.pos_decreasing[:diff]

        return start_weights



import argparse
import sys
from pathlib import Path

from immuneML.api.galaxy.Util import Util
from immuneML.data_model.bnp_util import write_yaml
from immuneML.util.PathBuilder import PathBuilder

from immuneML.ml_metrics.ClusteringMetric import INTERNAL_EVAL_METRICS, EXTERNAL_EVAL_METRICS


def build_labels(labels_str):
    labels = labels_str.split(",")
    return [label.strip().strip("'\"") for label in labels]


def parse_command_line_arguments(args):
    parser = argparse.ArgumentParser(
        description="Tool for building specification for applying previously trained ML models in Galaxy")

    parser.add_argument("-l", "--labels", type=str, default="",
                        help="Which metadata labels should be predicted for the dataset (separated by comma).")
    parser.add_argument("-e", "--eval_metrics", type=str, choices=INTERNAL_EVAL_METRICS + EXTERNAL_EVAL_METRICS,
                        default=[], nargs="+",
                        help="External evaluation metrics to use for clustering, for these metrics, clusters are compared to a provided label.")

    parser.add_argument("-k", "--encoding_k", type=int, required=True, help="")
    parser.add_argument("-n", "--n_clusters", type=int, required=True, help="")
    parser.add_argument("-d", "--dim_red_method", type=str, choices=["PCA", "UMAP", "TSNE", "None"], default="None",
                        help="External evaluation metrics to use for clustering, for these metrics, clusters are compared to a provided label.")
    parser.add_argument("-t", "--training_percentage", type=int, default=100)

    parser.add_argument("-o", "--output_path", required=True,
                        help="Output location for the generated yaml file (directory).")
    parser.add_argument("-f", "--file_name", default="specs.yaml",
                        help="Output file name for the yaml file. Default name is 'specs.yaml' if not specified.")

    return parser.parse_args(args)


def build_specs(parsed_args):
    specs = {
        "definitions": {
            "datasets": {
                "dataset": {
                    "format": "AIRR",
                    "params": {"dataset_file": Util.discover_dataset_path()}
                }
            },
            'reports': {
                'dim_reduction': {'DimensionalityReduction': {"dim_red_method": {}}}
            },
            'encodings': {
                'kmer': {
                    'KmerFrequency': {
                        "k": parsed_args.encoding_k
                    }
                }
            },
            'ml_methods': {
                'kmeans': {
                    'KMeans': {
                        'n_clusters': parsed_args.n_clusters
                    }
                },
                "pca_2": {
                    "PCA": {
                        "n_components": 2
                    }
                }
            }
        },
        "instructions": {
            f"clustering": {
                "type": "Clustering",
                "dataset": "dataset",
                'metrics': parsed_args.eval_metrics,
                'clustering_settings': [
                    {'encoding': 'kmer', "dim_reduction": "pca_2", 'method': 'kmeans'},
                ],
                'split_config': {
                    'split_strategy': 'random',
                    'training_percentage': parsed_args.training_percentage / 100
                }
            }
        }
    }

    if parsed_args.labels != "":
        specs["instructions"]["clustering"]["labels"] = build_labels(parsed_args.labels)
        specs['definitions']['reports']['dim_reduction']["DimensionalityReduction"]['label'] = (
            specs)["instructions"]["clustering"]["labels"][0]

    if parsed_args.dim_red_method in ["PCA", "TSNE", "UMAP"]:
        specs["definitions"]["ml_methods"][parsed_args.dim_red_method.lower()] = {
            parsed_args.dim_red_method: {"n_components": 2}}
        if parsed_args.dim_red_method == 'TSNE':
            specs["definitions"]["ml_methods"][parsed_args.dim_red_method.lower()][parsed_args.dim_red_method]['init'] = "random"
        specs["definitions"]["reports"]["dim_reduction"]["DimensionalityReduction"]["dim_red_method"] = {
            parsed_args.dim_red_method.upper(): {"n_components": 2}
        }
        specs["instructions"]["clustering"]["reports"] = ["dim_reduction"]

    return specs


def main(args):
    parsed_args = parse_command_line_arguments(args)
    specs = build_specs(parsed_args)

    PathBuilder.build(parsed_args.output_path)
    output_location = Path(parsed_args.output_path) / parsed_args.file_name

    write_yaml(output_location, specs)

    return str(output_location)


if __name__ == "__main__":
    main(sys.argv[1:])



# coding=utf-8
# Copyright 2021 The EleutherAI and HuggingFace Teams. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modified forward-pass implementation based on https://github.com/huggingface/transformers/blob/main/src/transformers/models/gptj/modeling_gptj.py
# Modified for ImmuneML. Original code from https://github.com/salesforce/progen

from typing import Tuple

import numpy as np

import torch
import torch.utils.checkpoint
from torch import nn
from torch.nn import CrossEntropyLoss
from transformers import GenerationMixin

from transformers.activations import ACT2FN
from transformers.modeling_outputs import BaseModelOutputWithPast, CausalLMOutputWithPast
from transformers.modeling_utils import PreTrainedModel
from transformers.utils import logging
from transformers.utils.model_parallel_utils import assert_device_map, get_device_map
from .ProGenConfig import ProGenConfig


logger = logging.get_logger(__name__)


def fixed_pos_embedding(x, seq_dim=1, seq_len=None):
    dim = x.shape[-1]
    if seq_len is None:
        seq_len = x.shape[seq_dim]
    inv_freq = 1.0 / (10000 ** (torch.arange(0, dim, 2) / dim))
    sinusoid_inp = torch.einsum("i , j -> i j", torch.arange(seq_len), inv_freq).to(x.device).float()
    return torch.sin(sinusoid_inp), torch.cos(sinusoid_inp)


def rotate_every_two(x):
    x1 = x[:, :, :, ::2]
    x2 = x[:, :, :, 1::2]
    x = torch.stack((-x2, x1), axis=-1)
    return x.flatten(-2)  # in einsum notation: rearrange(x, '... d j -> ... (d j)')


def apply_rotary_pos_emb(x, sincos, offset=0):
    sin, cos = map(lambda t: t[None, offset : x.shape[1] + offset, None, :].repeat_interleave(2, 3), sincos)
    # einsum notation for lambda t: repeat(t[offset:x.shape[1]+offset,:], "n d -> () n () (d j)", j=2)
    return (x * cos) + (rotate_every_two(x) * sin)


class ProGenAttention(nn.Module):
    # Shared attention mask buffers are deterministic and replicated across layers; skip them when exporting to avoid shared-tensor errors.
    _keys_to_ignore_on_save = ["bias", "masked_bias"]
    _keys_to_ignore_on_load_missing = ["bias", "masked_bias"]

    def __init__(self, config):
        super().__init__()

        max_positions = config.max_position_embeddings
        self.register_buffer(
            "bias",
            torch.tril(torch.ones((max_positions, max_positions), dtype=torch.bool)).view(
                1, 1, max_positions, max_positions
            ),
        )
        self.register_buffer("masked_bias", torch.tensor(-1e9))

        self.attn_dropout = nn.Dropout(config.attn_pdrop)
        self.resid_dropout = nn.Dropout(config.resid_pdrop)

        self.embed_dim = config.hidden_size
        self.num_attention_heads = config.num_attention_heads
        self.head_dim = self.embed_dim // self.num_attention_heads
        if self.head_dim * self.num_attention_heads != self.embed_dim:
            raise ValueError(
                f"embed_dim must be divisible by num_attention_heads (got `embed_dim`: {self.embed_dim} and `num_attention_heads`: {self.num_attention_heads})."
            )
        self.scale_attn = torch.sqrt(torch.tensor(self.head_dim, dtype=torch.float32)).to(torch.get_default_dtype())
        self.qkv_proj = nn.Linear(self.embed_dim, self.embed_dim * 3, bias=False)

        self.out_proj = nn.Linear(self.embed_dim, self.embed_dim, bias=False)
        self.rotary_dim = None
        if config.rotary_dim is not None:
            self.rotary_dim = config.rotary_dim

    def _split_heads(self, x, n_head, dim_head, mp_num):
        reshaped = x.reshape(x.shape[:-1] + (n_head//mp_num, dim_head))
        reshaped = reshaped.reshape(x.shape[:-2] + (-1, ) + reshaped.shape[-1:])
        return reshaped

    def _merge_heads(self, tensor, num_attention_heads, attn_head_size):
        """
        Merges attn_head_size dim and num_attn_heads dim into n_ctx
        """
        if len(tensor.shape) == 5:
            tensor = tensor.permute(0, 1, 3, 2, 4).contiguous()
        elif len(tensor.shape) == 4:
            tensor = tensor.permute(0, 2, 1, 3).contiguous()
        else:
            raise ValueError(f"Input tensor rank should be one of [4, 5], but is: {len(tensor.shape)}")
        new_shape = tensor.size()[:-2] + (num_attention_heads * attn_head_size,)
        return tensor.view(new_shape)

    def _attn(
        self,
        query,
        key,
        value,
        attention_mask=None,
        head_mask=None,
    ):

        # compute causal mask from causal mask buffer
        query_length, key_length = query.size(-2), key.size(-2)
        causal_mask = self.bias[:, :, key_length - query_length : key_length, :key_length]

        # Keep the attention weights computation in fp32 to avoid overflow issues
        query = query.to(torch.float32)
        key = key.to(torch.float32)

        attn_weights = torch.matmul(query, key.transpose(-1, -2))

        attn_weights = attn_weights / self.scale_attn
        attn_weights = torch.where(causal_mask, attn_weights, self.masked_bias.to(attn_weights.dtype))

        if attention_mask is not None:
            # Apply the attention mask
            attn_weights = attn_weights + attention_mask

        attn_weights = nn.Softmax(dim=-1)(attn_weights)
        attn_weights = attn_weights.to(value.dtype)
        attn_weights = self.attn_dropout(attn_weights)

        # Mask heads if we want to
        if head_mask is not None:
            attn_weights = attn_weights * head_mask

        attn_output = torch.matmul(attn_weights, value)

        return attn_output, attn_weights

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        layer_past=None,
        head_mask=None,
        use_cache=False,
        output_attentions=False,
    ):

        qkv = self.qkv_proj(hidden_states)
        # TODO(enijkamp): factor out number of logical TPU-v3/v4 cores or make forward pass agnostic
        # mp_num = 4
        mp_num = 8
        qkv_split = qkv.reshape(qkv.shape[:-1] + (mp_num, -1))

        local_dim = self.head_dim * self.num_attention_heads // mp_num
        query, value, key = torch.split(qkv_split, local_dim, dim=-1)
        query = self._split_heads(query, self.num_attention_heads, self.head_dim, mp_num=mp_num)
        key = self._split_heads(key, self.num_attention_heads, self.head_dim, mp_num=mp_num)

        value = self._split_heads(value, self.num_attention_heads, self.head_dim, mp_num=mp_num)
        value = value.permute(0, 2, 1, 3)

        seq_len = key.shape[1]
        offset = 0

        if layer_past is not None:
            offset = layer_past[0].shape[-2]
            seq_len += offset

        if self.rotary_dim is not None:
            k_rot = key[:, :, :, : self.rotary_dim]
            k_pass = key[:, :, :, self.rotary_dim :]

            q_rot = query[:, :, :, : self.rotary_dim]
            q_pass = query[:, :, :, self.rotary_dim :]

            sincos = fixed_pos_embedding(k_rot, 1, seq_len=seq_len)
            k_rot = apply_rotary_pos_emb(k_rot, sincos, offset=offset)
            q_rot = apply_rotary_pos_emb(q_rot, sincos, offset=offset)

            key = torch.cat([k_rot, k_pass], dim=-1)
            query = torch.cat([q_rot, q_pass], dim=-1)
        else:
            sincos = fixed_pos_embedding(key, 1, seq_len=seq_len)
            key = apply_rotary_pos_emb(key, sincos, offset=offset)
            query = apply_rotary_pos_emb(query, sincos, offset=offset)

        key = key.permute(0, 2, 1, 3)
        query = query.permute(0, 2, 1, 3)

        if layer_past is not None:
            past_key = layer_past[0]
            past_value = layer_past[1]
            key = torch.cat((past_key, key), dim=-2)
            value = torch.cat((past_value, value), dim=-2)

        if use_cache is True:
            present = (key, value)
        else:
            present = None

        # compute self-attention: V x Softmax(QK^T)
        attn_output, attn_weights = self._attn(query, key, value, attention_mask, head_mask)

        attn_output = self._merge_heads(attn_output, self.num_attention_heads, self.head_dim)

        attn_output = self.out_proj(attn_output)
        attn_output = self.resid_dropout(attn_output)

        outputs = (attn_output, present)
        if output_attentions:
            outputs += (attn_weights,)

        return outputs  # a, present, (attentions)


class ProGenMLP(nn.Module):
    def __init__(self, intermediate_size, config):  # in MLP: intermediate_size= 4 * embed_dim
        super().__init__()
        embed_dim = config.n_embd

        self.fc_in = nn.Linear(embed_dim, intermediate_size)
        self.fc_out = nn.Linear(intermediate_size, embed_dim)

        self.act = ACT2FN[config.activation_function]
        self.dropout = nn.Dropout(config.resid_pdrop)

    def forward(self, hidden_states):
        hidden_states = self.fc_in(hidden_states)
        hidden_states = self.act(hidden_states)
        hidden_states = self.fc_out(hidden_states)
        hidden_states = self.dropout(hidden_states)
        return hidden_states


class ProGenBlock(nn.Module):
    def __init__(self, config):
        super().__init__()
        inner_dim = config.n_inner if config.n_inner is not None else 4 * config.n_embd
        self.ln_1 = nn.LayerNorm(config.n_embd, eps=config.layer_norm_epsilon)
        self.attn = ProGenAttention(config)
        self.mlp = ProGenMLP(inner_dim, config)

    def forward(
        self,
        hidden_states,
        layer_past=None,
        attention_mask=None,
        head_mask=None,
        use_cache=False,
        output_attentions=False,
    ):
        residual = hidden_states
        hidden_states = self.ln_1(hidden_states)
        attn_outputs = self.attn(
            hidden_states,
            layer_past=layer_past,
            attention_mask=attention_mask,
            head_mask=head_mask,
            use_cache=use_cache,
            output_attentions=output_attentions,
        )
        attn_output = attn_outputs[0]  # output_attn: a, present, (attentions)
        outputs = attn_outputs[1:]

        feed_forward_hidden_states = self.mlp(hidden_states)
        hidden_states = attn_output + feed_forward_hidden_states + residual

        if use_cache:
            outputs = (hidden_states,) + outputs
        else:
            outputs = (hidden_states,) + outputs[1:]

        return outputs  # hidden_states, present, (attentions)


class ProGenPreTrainedModel(PreTrainedModel):
    """
    An abstract class to handle weights initialization and a simple interface for downloading and loading pretrained
    models.
    """

    config_class = ProGenConfig
    base_model_prefix = "transformer"
    is_parallelizable = True

    def __init__(self, *inputs, **kwargs):
        super().__init__(*inputs, **kwargs)

    def _init_weights(self, module):
        """Initialize the weights."""
        if isinstance(module, (nn.Linear,)):
            # Slightly different from Mesh Transformer JAX which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)


class ProGenModel(ProGenPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)

        self.embed_dim = config.n_embd
        self.vocab_size = config.vocab_size
        self.wte = nn.Embedding(config.vocab_size, self.embed_dim)
        self.drop = nn.Dropout(config.embd_pdrop)
        self.h = nn.ModuleList([ProGenBlock(config) for _ in range(config.n_layer)])
        self.ln_f = nn.LayerNorm(self.embed_dim, eps=config.layer_norm_epsilon)
        self.rotary_dim = min(config.rotary_dim, config.n_ctx // config.num_attention_heads)
        self.init_weights()

        # Model parallel
        self.model_parallel = False
        self.device_map = None


    def parallelize(self, device_map=None):
        # Check validity of device_map
        self.device_map = (
            get_device_map(len(self.h), range(torch.cuda.device_count())) if device_map is None else device_map
        )
        assert_device_map(self.device_map, len(self.h))
        self.model_parallel = True
        self.first_device = "cpu" if "cpu" in self.device_map.keys() else "cuda:" + str(min(self.device_map.keys()))
        self.last_device = "cuda:" + str(max(self.device_map.keys()))
        self.wte = self.wte.to(self.first_device)
        # Load onto devices
        for k, v in self.device_map.items():
            for block in v:
                cuda_device = "cuda:" + str(k)
                self.h[block] = self.h[block].to(cuda_device)
        # ln_f to last
        self.ln_f = self.ln_f.to(self.last_device)


    def deparallelize(self):
        self.model_parallel = False
        self.device_map = None
        self.first_device = "cpu"
        self.last_device = "cpu"
        self.wte = self.wte.to("cpu")
        for index in range(len(self.h)):
            self.h[index] = self.h[index].to("cpu")
        self.ln_f = self.ln_f.to("cpu")
        torch.cuda.empty_cache()

    def get_input_embeddings(self):
        return self.wte

    def set_input_embeddings(self, new_embeddings):
        self.wte = new_embeddings

    def forward(
        self,
        input_ids=None,
        past_key_values=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            input_shape = input_ids.size()
            input_ids = input_ids.view(-1, input_shape[-1])
            batch_size = input_ids.shape[0]
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
            batch_size = inputs_embeds.shape[0]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        device = input_ids.device if input_ids is not None else inputs_embeds.device

        if token_type_ids is not None:
            token_type_ids = token_type_ids.view(-1, input_shape[-1])

        if position_ids is not None:
            position_ids = position_ids.view(-1, input_shape[-1])

        if past_key_values is None:
            past_length = 0
            past_key_values = tuple([None] * len(self.h))
        else:
            past_length = past_key_values[0][0].size(-2)

        if position_ids is None:
            position_ids = torch.arange(past_length, input_shape[-1] + past_length, dtype=torch.long, device=device)
            position_ids = position_ids.unsqueeze(0).view(-1, input_shape[-1])

        # Attention mask.
        if attention_mask is not None:
            assert batch_size > 0, "batch_size has to be defined and > 0"
            attention_mask = attention_mask.view(batch_size, -1)
            # We create a 3D attention mask from a 2D tensor mask.
            # Sizes are [batch_size, 1, 1, to_seq_length]
            # So we can broadcast to [batch_size, num_heads, from_seq_length, to_seq_length]
            # this attention mask is more simple than the triangular masking of causal attention
            # used in OpenAI GPT, we just need to prepare the broadcast dimension here.
            attention_mask = attention_mask[:, None, None, :]

            # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
            # masked positions, this operation will create a tensor which is 0.0 for
            # positions we want to attend and -10000.0 for masked positions.
            # Since we are adding it to the raw scores before the softmax, this is
            # effectively the same as removing these entirely.
            attention_mask = attention_mask.to(dtype=self.dtype)  # fp16 compatibility
            attention_mask = (1.0 - attention_mask) * -10000.0

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x num_attention_heads x N x N
        # head_mask has shape n_layer x batch x num_attention_heads x N x N
        head_mask = self.get_head_mask(head_mask, self.config.n_layer)

        if inputs_embeds is None:
            inputs_embeds = self.wte(input_ids)

        hidden_states = inputs_embeds

        if token_type_ids is not None:
            token_type_embeds = self.wte(token_type_ids)
            hidden_states = hidden_states + token_type_embeds

        hidden_states = self.drop(hidden_states)

        output_shape = input_shape + (hidden_states.size(-1),)

        presents = () if use_cache else None
        all_self_attentions = () if output_attentions else None
        all_hidden_states = () if output_hidden_states else None
        for i, (block, layer_past) in enumerate(zip(self.h, past_key_values)):

            # Model parallel
            if self.model_parallel:
                torch.cuda.set_device(hidden_states.device)
                # Ensure layer_past is on same device as hidden_states (might not be correct)
                if layer_past is not None:
                    layer_past = tuple(past_state.to(hidden_states.device) for past_state in layer_past)
                # Ensure that attention_mask is always on the same device as hidden_states
                if attention_mask is not None:
                    attention_mask = attention_mask.to(hidden_states.device)
                if isinstance(head_mask, torch.Tensor):
                    head_mask = head_mask.to(hidden_states.device)
            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)

            if getattr(self.config, "gradient_checkpointing", False) and self.training:

                if use_cache:
                    logger.warning(
                        "`use_cache=True` is incompatible with `config.gradient_checkpointing=True`. Setting "
                        "`use_cache=False`..."
                    )
                    use_cache = False

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        # None for past_key_value
                        return module(*inputs, use_cache, output_attentions)

                    return custom_forward

                outputs = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(block),
                    hidden_states,
                    None,
                    attention_mask,
                    head_mask[i],
                )
            else:
                outputs = block(
                    hidden_states,
                    layer_past=layer_past,
                    attention_mask=attention_mask,
                    head_mask=head_mask[i],
                    use_cache=use_cache,
                    output_attentions=output_attentions,
                )

            hidden_states = outputs[0]
            if use_cache is True:
                presents = presents + (outputs[1],)

            if output_attentions:
                all_self_attentions = all_self_attentions + (outputs[2 if use_cache else 1],)

            # Model Parallel: If it's the last layer for that device, put things on the next device
            if self.model_parallel:
                for k, v in self.device_map.items():
                    if i == v[-1] and "cuda:" + str(k) != self.last_device:
                        hidden_states = hidden_states.to("cuda:" + str(k + 1))

        hidden_states = self.ln_f(hidden_states)

        hidden_states = hidden_states.view(*output_shape)
        # Add last hidden state
        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        if not return_dict:
            return tuple(v for v in [hidden_states, presents, all_hidden_states, all_self_attentions] if v is not None)

        return BaseModelOutputWithPast(
            last_hidden_state=hidden_states,
            past_key_values=presents,
            hidden_states=all_hidden_states,
            attentions=all_self_attentions,
        )


class ProGenForCausalLM(ProGenPreTrainedModel, GenerationMixin):
    _keys_to_ignore_on_load_missing = [r"h\.\d+\.attn\.masked_bias", r"h\.\d+\.attn\.bias", r"lm_head\.weight"]

    def __init__(self, config):
        super().__init__(config)
        self.transformer = ProGenModel(config)
        self.lm_head = nn.Linear(config.n_embd, config.vocab_size)
        self.init_weights()

        # Model parallel
        self.model_parallel = False
        self.device_map = None

    def parallelize(self, device_map=None):
        self.device_map = (
            get_device_map(len(self.transformer.h), range(torch.cuda.device_count()))
            if device_map is None
            else device_map
        )
        assert_device_map(self.device_map, len(self.transformer.h))
        self.transformer.parallelize(self.device_map)
        self.lm_head = self.lm_head.to(self.transformer.first_device)
        self.model_parallel = True

    def deparallelize(self):
        self.transformer.deparallelize()
        self.transformer = self.transformer.to("cpu")
        self.lm_head = self.lm_head.to("cpu")
        self.model_parallel = False
        torch.cuda.empty_cache()

    def get_output_embeddings(self):
        return None

    def set_output_embeddings(self, new_embeddings):
        return

    def prepare_inputs_for_generation(self, input_ids, past=None, **kwargs):
        token_type_ids = kwargs.get("token_type_ids", None)
        # only last token for inputs_ids if past is defined in kwargs
        if past:
            input_ids = input_ids[:, -1].unsqueeze(-1)
            if token_type_ids is not None:
                token_type_ids = token_type_ids[:, -1].unsqueeze(-1)

        attention_mask = kwargs.get("attention_mask", None)
        position_ids = kwargs.get("position_ids", None)

        if attention_mask is not None and position_ids is None:
            # create position_ids on the fly for batch generation
            position_ids = attention_mask.long().cumsum(-1) - 1
            position_ids.masked_fill_(attention_mask == 0, 1)
            if past:
                position_ids = position_ids[:, -1].unsqueeze(-1)
        else:
            position_ids = None
        return {
            "input_ids": input_ids,
            "past_key_values": past,
            "use_cache": kwargs.get("use_cache"),
            "position_ids": position_ids,
            "attention_mask": attention_mask,
            "token_type_ids": token_type_ids,
        }

    def forward(
        self,
        input_ids=None,
        past_key_values=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size, sequence_length)`, `optional`):
            Labels for language modeling. Note that the labels **are shifted** inside the model, i.e. you can set
            ``labels = input_ids`` Indices are selected in ``[-100, 0, ..., config.vocab_size]`` All labels set to
            ``-100`` are ignored (masked), the loss is only computed for labels in ``[0, ..., config.vocab_size]``
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        transformer_outputs = self.transformer(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        hidden_states = transformer_outputs[0]

        # Set device for model parallelism
        if self.model_parallel:
            torch.cuda.set_device(self.transformer.first_device)
            hidden_states = hidden_states.to(self.lm_head.weight.device)

        # make sure sampling in fp16 works correctly and
        # compute loss in fp32 to match with mesh-tf version
        # https://github.com/EleutherAI/gpt-neo/blob/89ce74164da2fb16179106f54e2269b5da8db333/models/gpt2/gpt2.py#L179
        lm_logits = self.lm_head(hidden_states).to(torch.float32)

        loss = None
        if labels is not None:
            # Shift so that tokens < n predict n
            shift_logits = lm_logits[..., :-1, :].contiguous()
            shift_labels = labels[..., 1:].contiguous()
            # Flatten the tokens
            loss_fct = CrossEntropyLoss()
            loss = loss_fct(shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1))

            loss = loss.to(hidden_states.dtype)

        if not return_dict:
            output = (lm_logits,) + transformer_outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return CausalLMOutputWithPast(
            loss=loss,
            logits=lm_logits,
            past_key_values=transformer_outputs.past_key_values,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
        )

    @staticmethod
    def _reorder_cache(past: Tuple[Tuple[torch.Tensor]], beam_idx: torch.Tensor) -> Tuple[Tuple[torch.Tensor]]:
        """
        This function is used to re-order the :obj:`past_key_values` cache if
        :meth:`~transformers.PretrainedModel.beam_search` or :meth:`~transformers.PretrainedModel.beam_sample` is
        called. This is required to match :obj:`past_key_values` with the correct beam_idx at every generation step.
        """
        return tuple(
            tuple(past_state.index_select(0, beam_idx.to(past_state.device)) for past_state in layer_past)
            for layer_past in past
        )



from pathlib import Path
from typing import List

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.IO.dataset_export.DataExporter import DataExporter
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.preprocessing.Preprocessor import Preprocessor
from immuneML.util.Logger import print_log
from immuneML.util.ReflectionHandler import ReflectionHandler
from immuneML.workflows.instructions.Instruction import Instruction
from immuneML.workflows.instructions.dataset_generation.DatasetExportState import DatasetExportState
from scripts.specification_util import update_docs_per_mapping


class DatasetExportInstruction(Instruction):
    """
    DatasetExport instruction takes a list of datasets as input, optionally applies preprocessing steps, and outputs
    the data in specified formats.

    **Specification arguments:**

    - datasets (list): a list of datasets to export in all given formats

    - preprocessing_sequence (list): which preprocessing sequence to use on the dataset(s), this item is optional and does not have to be specified.
      When specified, the same preprocessing sequence will be applied to all datasets.

    - number_of_processes (int): how many processes to use during repertoire export (not used for sequence datasets)

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        instructions:
            my_dataset_export_instruction: # user-defined instruction name
                type: DatasetExport # which instruction to execute
                datasets: # list of datasets to export
                    - my_generated_dataset
                    - my_dataset_from_adaptive
                preprocessing_sequence: my_preprocessing_sequence
                number_of_processes: 4

    """

    def __init__(self, datasets: List[Dataset], number_of_processes: int = 1,
                 preprocessing_sequence: List[Preprocessor] = None, result_path: Path = None, name: str = None):
        self.datasets = datasets
        self.preprocessing_sequence = preprocessing_sequence
        self.result_path = result_path
        self.number_of_processes = number_of_processes
        self.name = name

    def run(self, result_path: Path) -> DatasetExportState:
        self.result_path = result_path / self.name
        paths = {}
        export_format = 'AIRR'

        for dataset in self.datasets:
            dataset_name = dataset.name if dataset.name is not None else dataset.identifier

            if self.preprocessing_sequence is not None and len(self.preprocessing_sequence) > 0:
                for index, preprocessing in enumerate(self.preprocessing_sequence):
                    print_log(
                        f"For dataset {dataset_name}, started preprocessing step {index + 1}/{len(self.preprocessing_sequence)} with {preprocessing.__class__.__name__}",
                        include_datetime=True)
                    dataset = preprocessing.process_dataset(dataset, self.result_path / f"step_{index + 1}")
                    print_log(
                        f"Preprocessed {dataset.__class__.__name__.split('Dataset')[0].lower()} dataset {dataset.name} with {preprocessing.__class__.__name__}:\n"
                        f"- Example count: {dataset.get_example_count()}\n"
                        f"- Labels: {dataset.get_label_names()}", True)

            dataset.name = dataset_name

            paths[dataset_name] = {}
            path = self.result_path / dataset_name / export_format
            AIRRExporter.export(dataset, path, number_of_processes=self.number_of_processes)
            paths[dataset_name][export_format] = path
            contains = str(dataset.__class__.__name__).replace("Dataset", "s").lower()
            print_log(f"Exported dataset {dataset_name} containing {dataset.get_example_count()} "
                      f"{contains} in {export_format} format.", include_datetime=True)

        return DatasetExportState(datasets=self.datasets,
                                  formats=[export_format],
                                  preprocessing_sequence=self.preprocessing_sequence, paths=paths,
                                  result_path=self.result_path, name=self.name)

    @staticmethod
    def get_documentation():
        doc = str(DatasetExportInstruction.__doc__)

        valid_strategy_values = ReflectionHandler.all_nonabstract_subclass_basic_names(DataExporter, "Exporter",
                                                                                       "dataset_export/")
        valid_strategy_values = str(valid_strategy_values)[1:-1].replace("'", "`")
        mapping = {
            "Valid formats are class names of any non-abstract class inheriting "
            ":py:obj:`~immuneML.IO.dataset_export.DataExporter.DataExporter`.": f"Valid values are: {valid_strategy_values}.",
            "preprocessing_sequence (list)": "preprocessing_sequence (str)",
            "exporters (list)": "formats (list)"
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



import shutil
from pathlib import Path

import numpy as np
import pandas as pd
import torch
from plotly.subplots import make_subplots
import plotly.graph_objects as go
import plotly.express as px
from sklearn.decomposition import PCA

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.dsl.definition_parsers.MLParser import MLParser
from immuneML.ml_methods.dim_reduction.DimRedMethod import DimRedMethod
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.ml_methods.generative_models.SimpleVAE import SimpleVAE
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.gen_model_reports.GenModelReport import GenModelReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class VAESummary(GenModelReport):
    """
    This report provides the summary of the train SimpleVAE and shows the following:

    - plots of the latent space after applying a dimensionality reduction method to reduce the data to 2 dimensions,
      highlighted by V and J gene
    - plots the histogram for each latent dimension
    - plots loss per epoch

    **Specification arguments:**

    - dim_dist_cols (int): how many columns to use to plot the histograms of latent dimensions (either this or
      dim_dist_rows has to be set, or both)

    - dim_dist_rows (int): how many rows to use to plot the histogram of latent dimensions (either this or
      dim_dist_cols has to be set, or both)

    - dim_red_method (dict): which dimensionality reduction method to use along with its parameters; the method has to
      return 2 dimensions; see :ref:`**Dimensionality reduction methods**` for available options;
      default is PCA with 2 components

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_vae_summary:
                    VAESummary:
                        dim_dist_cols: 4
                        dim_dist_rows: None
                        dim_red_method:
                            PCA:
                                n_components: 2

    """
    @classmethod
    def build_object(cls, **kwargs):
        name = kwargs["name"] if "name" in kwargs else "VAESummary"

        ParameterValidator.assert_keys_present(list(kwargs.keys()), ['dim_dist_cols', 'dim_dist_rows', 'dim_red_method'],
                                               VAESummary.__name__, 'parameters')

        cls_name = list(kwargs['dim_red_method'].keys())[0]
        method = MLParser.parse_any_model("dim_red_method", kwargs['dim_red_method'], cls_name)[0]
        assert len(method.get_dimension_names()) == 2, \
            (f"{VAESummary.__name__}: dimensionality reduction method has to return 2 dimensions, got "
             f"{len(method.get_dimension_names())} instead.")

        assert ((kwargs['dim_dist_cols'] is not None and isinstance(kwargs['dim_dist_cols'], int)) or
                (kwargs['dim_dist_rows'] is not None and isinstance(kwargs['dim_dist_rows'], int))), \
            f"{cls.__name__}: at least one of dim_dist_cols or dim_dist_rows has to be set."

        return VAESummary(name=name, dim_dist_cols=kwargs['dim_dist_cols'], dim_dist_rows=kwargs['dim_dist_rows'],
                          dim_red_method=method)

    def __init__(self, dim_dist_rows: int, dim_dist_cols: int, dataset: Dataset = None, model: SimpleVAE = None,
                 result_path: Path = None, name: str = None, dim_red_method: DimRedMethod = None):
        super().__init__(dataset, model, result_path, name)
        self.dim_dist_rows = dim_dist_rows
        self.dim_dist_cols = dim_dist_cols
        self.dim_red_method = dim_red_method

    @property
    def dim_red_method_name(self) -> str:
        return self.dim_red_method.__class__.__name__

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)

        embeddings = self.get_embeddings()

        latent_space_table_out = self._prepare_latent_space(embeddings)
        train_progress_table_out = self._prepare_training_progress()

        latent_space_fig_out = self._safe_plot(output_written=True, plot_callable='_plot_latent_space',
                                               latent_space_table=latent_space_table_out)
        latent_dim_dist_fig_out = self._safe_plot(output_written=True, embeddings=embeddings,
                                                  plot_callable='_plot_latent_dimension_distributions')
        train_progress_fig_out = self._safe_plot(output_written=True, plot_callable='_plot_training_progress',
                                                 training_progress_table=train_progress_table_out)

        output_figures = [latent_space_fig_out, latent_dim_dist_fig_out, train_progress_fig_out]
        output_figures = [el for el in output_figures if el is not None]

        result = ReportResult(name=self.name, info='Summary of the fitted VAE model.', output_figures=output_figures,
                              output_tables=[latent_space_table_out, train_progress_table_out])
        return result

    def get_embeddings(self):

        data_loader = self.model.encode_dataset(self.dataset, None, self.dataset.get_example_count(), False)

        for data in data_loader:
            with torch.no_grad():
                cdr3_input, v_gene_input, j_gene_input = data
                embeddings = self.model.model.encoding_func(cdr3_input, v_gene_input, j_gene_input)

        return embeddings

    def _prepare_latent_space(self, embeddings) -> ReportOutput:
        embeddings = self.dim_red_method.fit_transform(design_matrix=embeddings.numpy())

        df = pd.DataFrame(data=embeddings, columns=self.dim_red_method.get_dimension_names())
        df['v_gene'] = [v_call.split("*")[0] for v_call in self.dataset.data.v_call.tolist()]
        df['j_gene'] = [j_call.split("*")[0] for j_call in self.dataset.data.j_call.tolist()]

        path = self.result_path / f'latent_space_{self.dim_red_method_name}.csv'

        df.to_csv(str(path), index=False)

        return ReportOutput(path, f'principal component analysis on the data embedded into '
                                  f'{self.model.latent_dim} dimensional space')

    def _prepare_training_progress(self) -> ReportOutput:
        path = self.result_path / 'training_losses.csv'
        shutil.copyfile(str(self.model.loss_path), str(path))
        return ReportOutput(path, 'Loss per epoch')

    def _plot_latent_space(self, latent_space_table: ReportOutput) -> ReportOutput:

        encoded_data = pd.read_csv(latent_space_table.path)
        dim_names = self.dim_red_method.get_dimension_names()

        fig = make_subplots(rows=2, cols=1, shared_xaxes=True, shared_yaxes=True, subplot_titles=['V genes', 'J genes'],
                            x_title=dim_names[0], y_title=dim_names[1])

        for gene in self.model.unique_v_genes:
            tmp_df = encoded_data[encoded_data['v_gene'] == gene]
            fig.add_trace(
                go.Scatter(x=tmp_df[dim_names[0]].values, y=tmp_df[dim_names[1]].values, mode='markers', opacity=0.5,
                           name=gene,
                           marker=dict(colorscale='Viridis'), legendgroup=1,
                           legendgrouptitle_text='V genes'), row=1, col=1)
        for gene in self.model.unique_j_genes:
            tmp_df = encoded_data[encoded_data['j_gene'] == gene]
            fig.add_trace(
                go.Scatter(x=tmp_df[dim_names[0]].values, y=tmp_df[dim_names[1]].values, mode='markers', opacity=0.5,
                           name=gene,
                           marker=dict(colorscale='Viridis'),
                           legendgroup=2, legendgrouptitle_text='J genes'), row=2, col=1)

        fig.update_layout(template='plotly_white', legend_tracegroupgap=30)
        file_path = PlotlyUtil.write_image_to_file(fig, self.result_path / f'latent_space_{self.dim_red_method_name}.html')

        return ReportOutput(file_path,
                            f'principal component analysis on the data embedded into {self.model.latent_dim} '
                            f'dimensional space')

    def _plot_latent_dimension_distributions(self, embeddings) -> ReportOutput:

        if self.dim_dist_rows is None:
            self.dim_dist_rows = int(self.model.latent_dim / self.dim_dist_cols)
        elif self.dim_dist_cols is None:
            self.dim_dist_cols = int(self.model.latent_dim / self.dim_dist_rows)

        assert self.dim_dist_rows * self.dim_dist_cols == self.model.latent_dim, \
            (f"{VAESummary.__name__}: cannot plot latent dimension distribution since dim_dist_cols and dim_dist_rows "
             f"parameters do not match the latent dimension of the VAE.")

        fig = make_subplots(rows=self.dim_dist_rows, cols=self.dim_dist_cols, shared_xaxes=True, shared_yaxes=True)

        i = 0
        for row in range(self.dim_dist_rows):
            for col in range(self.dim_dist_cols):
                fig.add_trace(go.Histogram(x=embeddings[:, i].numpy(), name=f'dim {i + 1}'),
                              row=row + 1, col=col + 1)
                i += 1

        fig.update_layout(template='plotly_white')
        file_path = PlotlyUtil.write_image_to_file(fig, self.result_path / 'latent_dim_dist.html')

        return ReportOutput(path=file_path,
                            name=f'latent dimension distribution for examples from dataset {self.dataset.name}')

    def _plot_training_progress(self, training_progress_table: ReportOutput) -> ReportOutput:
        df = pd.read_csv(str(training_progress_table.path))
        fig = px.line(df, x='epoch', y='loss', markers=True)
        fig.update_layout(template='plotly_white')
        file_path = PlotlyUtil.write_image_to_file(fig, self.result_path / 'loss_per_epoch.html')

        return ReportOutput(path=file_path, name='loss per epoch')

    def check_prerequisites(self) -> bool:
        return isinstance(self.model, SimpleVAE)



from enum import Enum
from typing import List, Union


class Chain(Enum):
    ALPHA = "TRA"
    BETA = "TRB"
    GAMMA = "TRG"
    DELTA = "TRD"
    HEAVY = "IGH"
    LIGHT = "IGL"
    KAPPA = "IGK"

    @staticmethod
    def get_chain(item: Union[str, 'Chain']):
        if type(item) is Chain:
            return item
        elif item.upper() in ["TRA", "TCRA", "A", "ALPHA", "TCRA"]:
            return Chain.ALPHA
        elif item.upper() in ["TRB", "TCRB", "B", "BETA", "TCRB"]:
            return Chain.BETA
        elif item.upper() in ["TRD", "TCRD", "D", "DELTA", "TCRD"]:
            return Chain.DELTA
        elif item.upper() in ["TRG", "TCRG", "G", "GAMMA", "TCRG"]:
            return Chain.GAMMA
        elif item.upper() in ["IGH", "H", "HEAVY"]:
            return Chain.HEAVY
        elif item.upper() in ["IGL", "L", "LIGHT", "LAMBDA"]:
            return Chain.LIGHT
        elif item.upper() in ["IGK", "K", "KAPPA"]:
            return Chain.KAPPA
        else:
            return None

    @staticmethod
    def get_chain_value(item: str):
        chain = Chain.get_chain(item)
        if chain:
            return chain.value
        else:
            return ''

    def __str__(self):
        return self.name

    def to_string(self):
        return self.value


class ChainPair(Enum):
    TRA_TRB = (Chain.ALPHA.value, Chain.BETA.value)
    TRG_TRD = (Chain.GAMMA.value, Chain.DELTA.value)
    IGH_IGL = (Chain.HEAVY.value, Chain.LIGHT.value)
    IGH_IGK = (Chain.HEAVY.value, Chain.KAPPA.value)

    @staticmethod
    def is_allowed(chain: Chain, chain_pair: 'ChainPair'):
        return chain.value in chain_pair.value

    @staticmethod
    def get_chain_pair(chains: List[Chain]):
        """Given a list of 2 chain objects, returns the relevant ChainPair"""
        assert len(chains) == 2, f"ChainPair: expected 2 chains, received {len(chains)}"
        assert type(chains[0]) == Chain and type(
            chains[1]) == Chain, f"ChainPair: expected Chain objects, received {type(chains[0])} and {type(chains[1])}"

        if Chain.ALPHA in chains and Chain.BETA in chains:
            return ChainPair.TRA_TRB
        elif Chain.GAMMA in chains and Chain.DELTA in chains:
            return ChainPair.TRG_TRD
        elif Chain.HEAVY in chains and Chain.LIGHT in chains:
            return ChainPair.IGH_IGL
        elif Chain.HEAVY in chains and Chain.KAPPA in chains:
            return ChainPair.IGH_IGK
        else:
            raise ValueError(f"ChainPair: illegal chain combination: {chains[0]} and {chains[1]}")


class RegionType(Enum):

    IMGT_CDR1 = "cdr1"
    IMGT_CDR2 = "cdr2"
    IMGT_CDR3 = "cdr3"
    IMGT_FR1 = "fwr1"
    IMGT_FR2 = "fwr2"
    IMGT_FR3 = "fwr3"
    IMGT_FR4 = "fwr4"
    IMGT_JUNCTION = "junction"
    FULL_SEQUENCE = "full_sequence"

    def to_string(self):
        return self.value.lower()

    @classmethod
    def get_object(cls, rt):
        if isinstance(rt, RegionType):
            return rt
        elif isinstance(rt, str):
            return RegionType[rt.upper()]
        else:
            raise RuntimeError(f"RegionType could not be created from {rt}.")




import logging
import pickle
from pathlib import Path

import numpy as np
from glmnet import LogitNet

from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.bnp_util import write_yaml
from immuneML.environment.Label import Label
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_methods.util.Util import Util
from immuneML.util.PathBuilder import PathBuilder


class LogRegressionCustomPenalty(MLMethod):
    """
    Logistic Regression with custom penalty factors for specific features.

    **Specification arguments**:

    - non_penalized_features (list): List of feature names that should not be penalized.

    - non_penalized_encodings (list): List of encoding names (class names) whose features should not be penalized. This
      parameter can be used only in combination with CompositeEncoder. None fo the features from the specified encodings
      will be penalized. If both non_penalized_features and non_penalized_encodings are provided, the union of the two
      will be used.

    Other supported arguments are inherited from LogitNet of python-glmnet package and will be directly passed to it.
    n_jobs will be overwritten to use the number of CPUs specified for the instruction (e.g. in TrainMLModel).

    **YAML specification:**

    .. code-block:: yaml

        ml_methods:
            custom_log_reg:
                LogRegressionCustomPenalty:
                    alpha: 1
                    n_lambda: 100
                    non_penalized_features: []
                    non_penalized_encodings: ['Metadata']
                    random_state: 42

    """
    def __init__(self, non_penalized_features: list = None, name: str = None, label: Label = None,
                 non_penalized_encodings: list = None, **kwargs):
        super().__init__(name=name, label=label)
        self.non_penalized_features = non_penalized_features if non_penalized_features is not None else []
        self.non_penalized_encodings = non_penalized_encodings if non_penalized_encodings is not None else []

        for ind, encoding in enumerate(self.non_penalized_encodings):
            if 'Encoder' not in encoding:
                self.non_penalized_encodings[ind] = encoding + 'Encoder'

        self.model = None
        self.feature_names = None
        self.kwargs = kwargs

    def _fit(self, encoded_data: EncodedData, cores_for_training: int = 2):
        X = encoded_data.examples
        y = Util.map_to_new_class_values(encoded_data.labels[self.label.name], self.class_mapping)

        self.feature_names = encoded_data.feature_names

        if encoded_data.encoding == 'CompositeEncoder' and self.non_penalized_encodings:
            features_from_non_penalized_encodings = encoded_data.feature_annotations[encoded_data.feature_annotations['encoder'].isin(self.non_penalized_encodings)]['feature'].tolist()
            non_penalized_features = list(set(features_from_non_penalized_encodings))
            non_penalized_features.extend(self.non_penalized_features)
            self.non_penalized_features = list(set(non_penalized_features))

            logging.info(f"{self.__class__.__name__}: inferred non-penalized features: {self.non_penalized_features}")

        # Create penalty factor vector
        penalty_factor = np.ones(X.shape[1])
        for idx, feature in enumerate(self.feature_names):
            if feature in self.non_penalized_features:
                penalty_factor[idx] = 0.0

        self.model = LogitNet(**self.kwargs, **{'n_jobs': cores_for_training})
        self.model.fit(X, y, relative_penalties=penalty_factor)

    def _predict(self, encoded_data: EncodedData):
        predictions = self.model.predict(encoded_data.examples)
        return {self.label.name: Util.map_to_old_class_values(np.array(predictions), self.class_mapping)}

    def _predict_proba(self, encoded_data: EncodedData):
        class_names = Util.map_to_old_class_values(self.model.classes_, self.class_mapping)
        probabilities = self.model.predict_proba(encoded_data.examples)
        return {self.label.name: {class_name: probabilities[:, i] for i, class_name in enumerate(class_names)}}

    def store(self, path: Path):
        PathBuilder.build(path)
        write_yaml(path / 'model.yaml', vars(self))
        with open(path / 'model.pkl', 'wb') as f:
            pickle.dump({
                'model': self.model,
                'non_penalized_features': self.non_penalized_features,
                'feature_names': self.feature_names
            }, f)

    def load(self, path: Path):
        with open(path / 'model.pkl', 'rb') as f:
            model = pickle.load(f)
            self.model = model['model']
            self.non_penalized_features = model['non_penalized_features']
            self.feature_names = model['feature_names']

    def get_params(self, for_refitting=False) -> dict:
        return {
            'penalty': self.model.penalty,
            'C': self.model.C,
            'random_state': self.model.random_state,
            'non_penalized_features': self.non_penalized_features
        }

    def can_predict_proba(self) -> bool:
        return True

    def can_fit_with_example_weights(self) -> bool:
        return False  # LogitNet does not support sample weights

    def get_compatible_encoders(self):
        from immuneML.encodings.composite_encoding.CompositeEncoder import CompositeEncoder
        return [CompositeEncoder]



# quality: gold
import abc
import hashlib
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.EncodedData import EncodedData
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.preprocessing.FeatureScaler import FeatureScaler
from immuneML.encodings.word2vec.model_creator.KmerPairModelCreator import KmerPairModelCreator
from immuneML.encodings.word2vec.model_creator.ModelType import ModelType
from immuneML.encodings.word2vec.model_creator.SequenceModelCreator import SequenceModelCreator
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.EncoderHelper import EncoderHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.ReflectionHandler import ReflectionHandler
from scripts.specification_util import update_docs_per_mapping


class Word2VecEncoder(DatasetEncoder):
    """

    Word2VecEncoder learns the vector representations of k-mers based on the context (receptor sequence).
    Similar idea was discussed in: Ostrovsky-Berman, M., Frankel, B., Polak, P. & Yaari, G.
    Immune2vec: Embedding B/T Cell Receptor Sequences in ℝN Using Natural Language Processing. Frontiers in Immunology 12, (2021).

    This encoder relies on gensim's implementation of Word2Vec and KmerHelper for k-mer extraction. Currently it works on amino acid level.

    **Dataset type:**

    - SequenceDatasets

    - RepertoireDatasets


    **Specification arguments:**

    - vector_size (int): The size of the vector to be learnt.

    - model_type (:py:obj:`~immuneML.encodings.word2vec.model_creator.ModelType.ModelType`):  The context which will be
      used to infer the representation of the sequence.
      If :py:obj:`~immuneML.encodings.word2vec.model_creator.ModelType.ModelType.SEQUENCE` is used, the context of
      a k-mer is defined by the sequence it occurs in (e.g. if the sequence is CASTTY and k-mer is AST,
      then its context consists of k-mers CAS, STT, TTY)
      If :py:obj:`~immuneML.encodings.word2vec.model_creator.ModelType.ModelType.KMER_PAIR` is used, the context for
      the k-mer is defined as all the k-mers that within one edit distance (e.g. for k-mer CAS, the context
      includes CAA, CAC, CAD etc.).
      Valid values for this parameter are names of the ModelType enum.

    - k (int): The length of the k-mers used for the encoding.

    - epochs (int): for how many epochs to train the word2vec model for a given set of sentences (corresponding to epochs parameter in gensim package)

    - window (int): max distance between two k-mers in a sequence (same as window parameter in gensim's word2vec)


    **YAML pecification:**

    .. highlight:: yaml
    .. code-block:: yaml

        definitions:
            encodings:
                encodings:
                    my_w2v:
                        Word2Vec:
                            vector_size: 16
                            k: 3
                            model_type: SEQUENCE
                            epochs: 100
                            window: 8

    """

    DESCRIPTION_REPERTOIRES = "repertoires"
    DESCRIPTION_LABELS = "labels"

    dataset_mapping = {
        "RepertoireDataset": "W2VRepertoireEncoder",
        "SequenceDataset": "W2VSequenceEncoder"
    }

    def __init__(self, vector_size: int, k: int, model_type: ModelType, epochs: int, window: int, name: str = None):
        self.vector_size = vector_size
        self.k = k
        self.epochs = epochs
        self.window = window
        self.model_type = model_type
        self.model_path = None
        self.scaler = None
        self.name = name

    @staticmethod
    def _prepare_parameters(vector_size: int, k: int, model_type: str, epochs: int, window: int, name: str = None):
        location = "Word2VecEncoder"
        ParameterValidator.assert_type_and_value(vector_size, int, location, "vector_size", min_inclusive=1)
        ParameterValidator.assert_type_and_value(k, int, location, "k", min_inclusive=1)
        ParameterValidator.assert_in_valid_list(model_type.upper(), [item.name for item in ModelType], location, "model_type")
        ParameterValidator.assert_type_and_value(epochs, int, location, 'epochs', min_inclusive=1)
        ParameterValidator.assert_type_and_value(window, int, location, 'window')
        return {"vector_size": vector_size, "k": k, "model_type": ModelType[model_type.upper()], "name": name, "epochs": epochs, "window": window}

    @staticmethod
    def build_object(dataset=None, **params):
        EncoderHelper.check_dataset_type_available_in_mapping(dataset, Word2VecEncoder)

        prepared_params = Word2VecEncoder._prepare_parameters(**params)
        encoder = ReflectionHandler.get_class_by_name(
                Word2VecEncoder.dataset_mapping[dataset.__class__.__name__], "word2vec/")(**prepared_params)

        return encoder

    def encode(self, dataset, params: EncoderParams):
        params.model = vars(self)
        cache_params = self._prepare_caching_params(dataset, params)
        encoded_dataset = CacheHandler.memo_by_params(cache_params, lambda: self._encode_new_dataset(dataset, params))

        EncoderHelper.sync_encoder_with_cache(cache_params, lambda: {'model_path': self.model_path, 'scaler': self.scaler}, self,
                                              ['model_path', 'scaler'])

        return encoded_dataset

    def _prepare_caching_params(self, dataset, params: EncoderParams, vectors=None, description: str = ""):
        return (("dataset_id", dataset.identifier),
                ("example_ids", tuple(dataset.get_example_ids())),
                ("dataset_type", dataset.__class__.__name__),
                ("labels", tuple(params.label_config.get_labels_by_name())),
                ("vectors", hashlib.sha256(str(vectors).encode("utf-8")).hexdigest()),
                ("description", description),
                ("encoding", Word2VecEncoder.__name__),
                ("learn_model", params.learn_model),
                ("encoding_params", tuple([(key, getattr(self, key)) for key in vars(self)])), )

    def _encode_new_dataset(self, dataset, params: EncoderParams):
        if params.learn_model is True and not self._exists_model(params):
            model = self._create_model(dataset=dataset, params=params)
        else:
            model = self._load_model(params)

        vectors = model.wv
        del model

        encoded_dataset = self._encode_by_model(dataset, params, vectors)

        return encoded_dataset

    @abc.abstractmethod
    def _encode_labels(self, dataset, params: EncoderParams):
        pass

    def _encode_by_model(self, dataset, params: EncoderParams, vectors):
        examples = CacheHandler.memo_by_params(self._prepare_caching_params(dataset, params, vectors,
                                                                            Word2VecEncoder.DESCRIPTION_REPERTOIRES),
                                               lambda: self._encode_examples(dataset, vectors, params))

        if params.encode_labels:
            labels = CacheHandler.memo_by_params(self._prepare_caching_params(dataset, params, vectors, Word2VecEncoder.DESCRIPTION_LABELS),
                                                 lambda: self._encode_labels(dataset, params))
        else:
            labels = None

        if params.learn_model:
            self.scaler = StandardScaler(with_std=True, with_mean=True)
            scaled_examples = FeatureScaler.standard_scale_fit(self.scaler, examples)
        else:
            scaled_examples = FeatureScaler.standard_scale(self.scaler, examples)

        encoded_dataset = self._build_encoded_dataset(dataset, scaled_examples, labels, params)
        return encoded_dataset

    def _build_encoded_dataset(self, dataset: Dataset, scaled_examples, labels, params: EncoderParams):

        encoded_dataset = dataset.clone()

        label_names = params.label_config.get_labels_by_name()
        feature_names = [str(i) for i in range(scaled_examples.shape[1])]
        feature_annotations = pd.DataFrame({"feature": feature_names})

        encoded_dataset.encoded_data = EncodedData(examples=scaled_examples,
                                                   labels={label: labels[i] for i, label in
                                                           enumerate(label_names)} if labels is not None else None,
                                                   example_ids=dataset.get_example_ids(),
                                                   feature_names=feature_names,
                                                   feature_annotations=feature_annotations,
                                                   encoding=Word2VecEncoder.__name__,
                                                   info={'sequence_type': params.sequence_type,
                                                         'region_type': params.region_type})
        return encoded_dataset

    def _encode_examples(self, encoded_dataset, vectors, params):
        examples = np.zeros(shape=[encoded_dataset.get_example_count(), vectors.vector_size])
        for (index, example) in enumerate(encoded_dataset.get_data()):
            examples[index] = self._encode_item(example, vectors, params)
        return examples

    @abc.abstractmethod
    def _encode_item(self, example, vectors, params: EncoderParams):
        pass

    def _load_model(self, params):
        from gensim.models import Word2Vec

        self.model_path = self._create_model_path(params) if self.model_path is None else self.model_path
        model = Word2Vec.load(str(self.model_path))
        return model

    def _create_model(self, dataset, params: EncoderParams):

        if self.model_path is None:
            self.model_path = self._create_model_path(params)

        if self.model_type == ModelType.SEQUENCE:
            model_creator = SequenceModelCreator(epochs=self.epochs, window=self.window)
        else:
            model_creator = KmerPairModelCreator(epochs=self.epochs, window=self.window)

        PathBuilder.build(self.model_path.parent)
        model = model_creator.create_model(dataset=dataset,
                                           k=self.k,
                                           vector_size=self.vector_size,
                                           batch_size=params.pool_size,
                                           model_path=self.model_path,
                                           sequence_type=params.model.get('sequence_type', EnvironmentSettings.sequence_type))

        return model

    def _exists_model(self, params: EncoderParams) -> bool:
        return self._create_model_path(params).is_file()

    def _create_model_path(self, params: EncoderParams) -> Path:
        if self.model_path is None:
            return params.result_path / "W2V.model"
        else:
            return self.model_path

    def get_additional_files(self) -> List[str]:
        return [self.model_path]

    @staticmethod
    def load_encoder(encoder_file: Path):
        encoder = DatasetEncoder.load_encoder(encoder_file)
        encoder = DatasetEncoder.load_attribute(encoder, encoder_file, "model_path")
        return encoder

    @staticmethod
    def get_documentation():
        doc = str(Word2VecEncoder.__doc__)

        valid_values = str([model_type.name for model_type in ModelType])[1:-1].replace("'", "`")
        mapping = {
            "Valid values for this parameter are names of the ModelType enum.": f"Valid values are {valid_values}."
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc
